"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.api import backend as backend
from synalinks.api import callbacks as callbacks
from synalinks.api import config as config
from synalinks.api import datasets as datasets
from synalinks.api import embedding_models as embedding_models
from synalinks.api import hooks as hooks
from synalinks.api import initializers as initializers
from synalinks.api import language_models as language_models
from synalinks.api import metrics as metrics
from synalinks.api import modules as modules
from synalinks.api import ops as ops
from synalinks.api import optimizers as optimizers
from synalinks.api import programs as programs
from synalinks.api import rewards as rewards
from synalinks.api import saving as saving
from synalinks.api import tree as tree
from synalinks.api import utils as utils
from synalinks.src.backend import DataModel as DataModel
from synalinks.src.backend import name_scope as name_scope
from synalinks.src.backend.common.global_state import clear_session as clear_session
from synalinks.src.backend.common.json_data_model import JsonDataModel as JsonDataModel
from synalinks.src.backend.common.stateless_scope import StatelessScope as StatelessScope
from synalinks.src.backend.common.symbolic_data_model import (
    SymbolicDataModel as SymbolicDataModel,
)
from synalinks.src.backend.common.symbolic_scope import SymbolicScope as SymbolicScope
from synalinks.src.backend.config import api_base as api_base
from synalinks.src.backend.config import disable_telemetry as disable_telemetry
from synalinks.src.backend.config import enable_logging as enable_logging
from synalinks.src.backend.config import enable_observability as enable_observability
from synalinks.src.backend.config import get_seed as get_seed
from synalinks.src.backend.config import (
    is_observability_enabled as is_observability_enabled,
)
from synalinks.src.backend.config import is_telemetry_enabled as is_telemetry_enabled
from synalinks.src.backend.config import set_api_base as set_api_base
from synalinks.src.backend.config import set_seed as set_seed
from synalinks.src.backend.config import synalinks_home as synalinks_home
from synalinks.src.backend.pydantic.base import ChatMessage as ChatMessage
from synalinks.src.backend.pydantic.base import ChatMessages as ChatMessages
from synalinks.src.backend.pydantic.base import ChatRole as ChatRole
from synalinks.src.backend.pydantic.base import Embeddings as Embeddings
from synalinks.src.backend.pydantic.base import GenericInputs as GenericInputs
from synalinks.src.backend.pydantic.base import GenericIO as GenericIO
from synalinks.src.backend.pydantic.base import GenericOutputs as GenericOutputs
from synalinks.src.backend.pydantic.base import GenericResult as GenericResult
from synalinks.src.backend.pydantic.base import Instructions as Instructions
from synalinks.src.backend.pydantic.base import Prediction as Prediction
from synalinks.src.backend.pydantic.base import Score as Score
from synalinks.src.backend.pydantic.base import SimilaritySearch as SimilaritySearch
from synalinks.src.backend.pydantic.base import Stamp as Stamp
from synalinks.src.backend.pydantic.base import ToolCall as ToolCall
from synalinks.src.backend.pydantic.base import ToolCall as ToolCalling
from synalinks.src.backend.pydantic.base import Trainable as Trainable
from synalinks.src.backend.pydantic.base import TripletSearch as TripletSearch
from synalinks.src.backend.pydantic.base import is_chat_message as is_chat_message
from synalinks.src.backend.pydantic.base import is_chat_messages as is_chat_messages
from synalinks.src.backend.pydantic.base import is_embedded as is_embedded
from synalinks.src.backend.pydantic.base import is_embedding as is_embedding
from synalinks.src.backend.pydantic.base import is_embeddings as is_embeddings
from synalinks.src.backend.pydantic.base import is_instructions as is_instructions
from synalinks.src.backend.pydantic.base import is_prediction as is_prediction
from synalinks.src.backend.pydantic.base import (
    is_similarity_search as is_similarity_search,
)
from synalinks.src.backend.pydantic.base import is_stamped as is_stamped
from synalinks.src.backend.pydantic.base import is_tool_call as is_tool_call
from synalinks.src.backend.pydantic.base import is_trainable as is_trainable
from synalinks.src.backend.pydantic.base import is_triplet_search as is_triplet_search
from synalinks.src.embedding_models.embedding_model import (
    EmbeddingModel as EmbeddingModel,
)
from synalinks.src.initializers.initializer import Initializer as Initializer
from synalinks.src.knowledge_bases.knowledge_base import KnowledgeBase as KnowledgeBase
from synalinks.src.language_models.language_model import LanguageModel as LanguageModel
from synalinks.src.metrics.metric import Metric as Metric
from synalinks.src.modules.agents.function_calling_agent import (
    FunctionCallingAgent as FunctionCallingAgent,
)
from synalinks.src.modules.core.action import Action as Action
from synalinks.src.modules.core.branch import Branch as Branch
from synalinks.src.modules.core.decision import Decision as Decision
from synalinks.src.modules.core.generator import Generator as Generator
from synalinks.src.modules.core.generator import (
    default_prompt_template as default_prompt_template,
)
from synalinks.src.modules.core.identity import Identity as Identity
from synalinks.src.modules.core.input_module import Input as Input
from synalinks.src.modules.core.not_module import Not as Not
from synalinks.src.modules.core.tool import Tool as Tool
from synalinks.src.modules.knowledge.embed_knowledge import (
    EmbedKnowledge as EmbedKnowledge,
)
from synalinks.src.modules.knowledge.retrieve_knowledge import (
    RetrieveKnowledge as RetrieveKnowledge,
)
from synalinks.src.modules.knowledge.stamp_knowledge import (
    StampKnowledge as StampKnowledge,
)
from synalinks.src.modules.knowledge.update_knowledge import (
    UpdateKnowledge as UpdateKnowledge,
)
from synalinks.src.modules.masking.in_mask import InMask as InMask
from synalinks.src.modules.masking.out_mask import OutMask as OutMask
from synalinks.src.modules.merging.concat import Concat as Concat
from synalinks.src.modules.merging.concat import Concat as Concatenate
from synalinks.src.modules.merging.logical_and import And as And
from synalinks.src.modules.merging.logical_or import Or as Or
from synalinks.src.modules.merging.logical_xor import Xor as Xor
from synalinks.src.modules.module import Module as Module
from synalinks.src.modules.synthesis.python_synthesis import (
    PythonSynthesis as PythonSynthesis,
)
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought as ChainOfThought
from synalinks.src.modules.ttc.self_critique import SelfCritique as SelfCritique
from synalinks.src.ops.function import Function as Function
from synalinks.src.ops.operation import Operation as Operation
from synalinks.src.programs.program import Program as Program
from synalinks.src.programs.sequential import Sequential as Sequential
from synalinks.src.rewards.cosine_similarity import CosineSimilarity as CosineSimilarity
from synalinks.src.rewards.exact_match import ExactMatch as ExactMatch
from synalinks.src.rewards.lm_as_judge import LMAsJudge as LMAsJudge
from synalinks.src.rewards.reward import Reward as Reward
from synalinks.src.rewards.reward_wrappers import ProgramAsJudge as ProgramAsJudge
from synalinks.src.utils.mcp.client import MultiServerMCPClient as MultiServerMCPClient
from synalinks.src.version import __version__
from synalinks.src.version import version as version



from synalinks.src.optimizers.evolutionary_optimizer import EvolutionaryOptimizer
from synalinks.src.optimizers.greedy_optimizer import GreedyOptimizer
from synalinks.src.optimizers.omega import OMEGA
from synalinks.src.optimizers.optimizer import Optimizer
from synalinks.src.optimizers.random_few_shot import RandomFewShot



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import ChatMessage
from synalinks.src.backend import ChatMessages
from synalinks.src.backend import ChatRole
from synalinks.src.backend import DataModel
from synalinks.src.language_models import LanguageModel


class LanguageModelTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_call_api_without_structured_output(self, mock_completion):
        language_model = LanguageModel(model="ollama/mistral")

        messages = ChatMessages(
            messages=[ChatMessage(role=ChatRole.USER, content="Hello")]
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": "Hello, how can I help you?"}}]
        }

        expected = ChatMessage(
            role=ChatRole.ASSISTANT, content="Hello, how can I help you?"
        )
        result = await language_model(messages)
        self.assertEqual(result, ChatMessage(**result).get_json())
        self.assertEqual(result, expected.get_json())

    @patch("litellm.acompletion")
    async def test_call_api_with_structured_output(self, mock_completion):
        language_model = LanguageModel(model="ollama/mistral")

        messages = ChatMessages(
            messages=[
                ChatMessage(
                    role=ChatRole.USER,
                    content="What is the french city of aerospace and robotics?",
                )
            ]
        )

        expected_string = (
            """{"rationale": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        expected = AnswerWithRationale(
            rationale=(
                "Toulouse hosts numerous research institutions and universities "
                "that specialize in aerospace engineering and robotics, such as "
                "the Institut Supérieur de l'Aéronautique et de l'Espace "
                "(ISAE-SUPAERO) and the "
                "French National Centre for Scientific Research (CNRS)"
            ),
            answer="Toulouse",
        )
        result = await language_model(messages, schema=AnswerWithRationale.get_schema())
        self.assertEqual(result, AnswerWithRationale(**result).get_json())
        self.assertEqual(result, expected.get_json())

    @patch("litellm.acompletion")
    async def test_call_api_streaming_mode(self, mock_completion):
        language_model = LanguageModel(model="ollama/deepseek-r1")

        messages = ChatMessages(
            messages=[ChatMessage(role=ChatRole.USER, content="Hello")]
        )

        mock_response_iterator = iter(
            [
                {"choices": [{"delta": {"content": "Hel"}}]},
                {"choices": [{"delta": {"content": "lo,"}}]},
                {"choices": [{"delta": {"content": " how"}}]},
                {"choices": [{"delta": {"content": " can"}}]},
                {"choices": [{"delta": {"content": " I"}}]},
                {"choices": [{"delta": {"content": " help"}}]},
                {"choices": [{"delta": {"content": " you?"}}]},
            ]
        )

        mock_completion.return_value = mock_response_iterator

        expected = "Hello, how can I help you?"

        response = await language_model(messages, streaming=True)

        result = ""
        for msg in response:
            result += msg.get("content")

        self.assertEqual(result, expected)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union


class DatabaseAdapter:
    """Base class for database adapters.

    DatabaseAdapter provides a unified interface for storing and retrieving
    structured data with optional embedding-based similarity search capabilities.

    Subclasses must implement the abstract methods to provide concrete
    database functionality.
    """

    def __init__(
        self,
        uri=None,
        embedding_model=None,
        data_models=None,
        metric="cosine",
        wipe_on_start=False,
        name=None,
        **kwargs,
    ):
        """Initialize the database adapter.

        Args:
            uri (str): The database connection URI or path.
            embedding_model (EmbeddingModel): Optional embedding model for
                vector similarity search.
            data_models (list): Optional list of SymbolicDataModel or DataModel
                classes to create tables for.
            metric (str): Distance metric for vector search. Options depend on
                the specific adapter implementation.
            wipe_on_start (bool): Whether to clear the database on initialization.
            name (str): Optional name for the adapter instance.
        """
        self.uri = uri
        self.embedding_model = embedding_model
        self.data_models = data_models or []
        self.metric = metric
        self.name = name

        if wipe_on_start:
            self.wipe_database()

    def wipe_database(self):
        """Clear all data from the database.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `wipe_database()` method"
        )

    def get_symbolic_data_models(self):
        """Retrieve all data models from the database schema.

        Returns:
            List[SymbolicDataModel]: List of symbolic data models representing
                the database schema.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the "
            "`get_symbolic_data_models()` method"
        )

    async def update(
        self,
        data_model_or_data_models: Union[Any, List[Any]],
    ) -> Union[Any, List[Any]]:
        """Insert or update records in the database.

        Args:
            data_model_or_data_models: A single JsonDataModel or a list of
                JsonDataModels to insert or update.

        Returns:
            The primary key value(s) of the inserted/updated records.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `update()` method"
        )

    async def get(
        self,
        id_or_ids: Any,
        data_models: Optional[List[Any]] = None,
    ) -> Optional[Any]:
        """Retrieve a record by its primary key.

        Args:
            id_or_ids: The primary key value to look up.
            data_models: Optional list of SymbolicDataModels to search in.
                If not provided, searches all tables.

        Returns:
            JsonDataModel if found, None otherwise.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `get()` method"
        )

    async def getall(
        self,
        data_model: Any,
        limit: int = 50,
        offset: int = 0,
    ) -> List[Any]:
        """Retrieve all records from a table with pagination.

        Args:
            data_model: The SymbolicDataModel representing the table to query.
            limit: Maximum number of records to return.
            offset: Number of records to skip.

        Returns:
            List of JsonDataModels.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `getall()` method"
        )

    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """Execute a raw query against the database.

        Args:
            query: The query string (SQL or other query language).
            params: Optional parameters for parameterized queries.
            **kwargs: Additional adapter-specific options.

        Returns:
            List of result dictionaries.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `query()` method"
        )

    async def similarity_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[Any]] = None,
        k: int = 10,
        threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Perform vector similarity search using embeddings.

        Requires an embedding_model to be configured.

        Args:
            text_or_texts: Query text or list of query texts to search for.
            data_models: Optional list of SymbolicDataModels to search in.
            k: Maximum number of results to return.
            threshold: Optional similarity threshold for filtering results.

        Returns:
            List of matching records with similarity scores.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `similarity_search()` method"
        )

    async def fulltext_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[Any]] = None,
        k: int = 10,
        threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Perform full-text search on text fields.

        Args:
            text_or_texts: Query text or list of query texts to search for.
            data_models: Optional list of SymbolicDataModels to search in.
            k: Maximum number of results to return.
            threshold: Optional relevance threshold for filtering results.

        Returns:
            List of matching records with relevance scores.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `fulltext_search()` method"
        )

    async def hybrid_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[Any]] = None,
        k: int = 10,
        similarity_threshold: Optional[float] = None,
        fulltext_threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Perform hybrid search combining vector similarity and full-text search.

        Uses Reciprocal Rank Fusion (RRF) to combine results from both
        similarity search and full-text search.

        Args:
            text_or_texts: Query text or list of query texts to search for.
            data_models: Optional list of SymbolicDataModels to search in.
            k: Maximum number of results to return.
            similarity_threshold: Optional threshold for vector similarity.
            fulltext_threshold: Optional threshold for full-text relevance.

        Returns:
            List of matching records with combined scores.

        Raises:
            NotImplementedError: Subclasses must implement this method.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} should implement the `hybrid_search()` method"
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} uri={self.uri}>"



# Modified from: keras/src/metrics/metric.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import backend
from synalinks.src import initializers
from synalinks.src.api_export import synalinks_export
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable
from synalinks.src.utils.naming import auto_name
from synalinks.src.utils.tracking import Tracker


@synalinks_export(["synalinks.Metric", "synalinks.metrics.Metric"])
class Metric(SynalinksSaveable):
    """Metric base class: all synalinks metrics inherit from this class.

    Args:
        name (str): (Optional) string name of the metric instance.
        in_mask (list): (Optional) list of keys to keep to compute the metric.
        out_mask (list): (Optional) list of keys to remove to compute the metric.
    """

    def __init__(self, name=None, in_mask=None, out_mask=None):
        self.name = name or auto_name(self.__class__.__name__)
        self._metrics = []
        self._variables = []
        self.in_mask = in_mask
        self.out_mask = out_mask
        self._tracker = Tracker(
            {
                "variables": (
                    lambda x: isinstance(x, backend.Variable),
                    self._variables,
                ),
                "metrics": (lambda x: isinstance(x, Metric), self._metrics),
            }
        )

    def reset_state(self):
        """Reset all of the metric state variables.

        This function is called between epochs/steps,
        when a metric is evaluated during training.
        """
        for v in self.variables:
            initializer = initializers.Empty(data_model=v._data_model)
            v.assign(initializer.get_json())

    async def update_state(self, *args, **kwargs):
        """Accumulate statistics for the metric."""
        raise NotImplementedError

    def stateless_update_state(self, metric_variables, *args, **kwargs):
        if len(metric_variables) != len(self.variables):
            raise ValueError(
                "Argument `metric_variables` must be a list of data_models "
                f"corresponding 1:1 to {self.__class__.__name__}().variables. "
                f"Received list with length {len(metric_variables)}, but "
                f"expected {len(self.variables)} variables."
            )
        # Gather variable mapping
        mapping = list(zip(self.variables, metric_variables))

        # Call in stateless scope
        with backend.StatelessScope(state_mapping=mapping) as scope:
            self.update_state(*args, **kwargs)

        # Gather updated variables
        metric_variables = []
        for v in self.variables:
            new_v = scope.get_current_value(v)
            if new_v is not None:
                metric_variables.append(new_v)
            else:
                metric_variables.append(v)
        return metric_variables

    def result(self):
        """Compute the current metric value.

        Returns:
            (float | dict): A scalar, or a dictionary of scalars.
        """
        raise NotImplementedError

    def stateless_result(self, metric_variables):
        if len(metric_variables) != len(self.variables):
            raise ValueError(
                "Argument `metric_variables` must be a list of data_models "
                f"corresponding 1:1 to {self.__class__.__name__}().variables. "
                f"Received list with length {len(metric_variables)}, but "
                f"expected {len(self.variables)} variables."
            )
        # Gather variable mapping
        mapping = list(zip(self.variables, metric_variables))

        # Call in stateless scope
        with backend.StatelessScope(state_mapping=mapping):
            res = self.result()
        return res

    def _obj_type(self):
        return "Metric"

    def add_variable(self, initializer=None, data_model=None, name=None):
        if initializer is None:
            initializer = initializers.Empty(data_model=data_model)
        self._check_super_called()
        with backend.name_scope(self.name.replace("/", ">"), caller=self):
            initializer = initializer
            variable = backend.Variable(
                initializer=initializer,
                data_model=data_model,
                trainable=False,
                name=name,
            )
        # Prevent double-tracking
        self._tracker.add_to_store("variables", variable)
        return variable

    @property
    def variables(self):
        variables = list(self._variables)
        for metric in self._metrics:
            variables.extend(metric.variables)
        return variables

    async def __call__(self, *args, **kwargs):
        self._check_super_called()
        await self.update_state(*args, **kwargs)
        return self.result()

    def get_config(self):
        """Return the serializable config of the metric.

        Returns:
            (dict): The config dict.
        """
        return {
            "in_mask": self.in_mask,
            "out_mask": self.out_mask,
            "name": self.name,
        }

    @classmethod
    def from_config(cls, config):
        """Returns a metric instance from config.

        Args:
            config (dict): The config dict.

        Returns:
            (Metric): The metric instance.
        """
        return cls(**config)

    def __setattr__(self, name, value):
        # Track Variables, Layers, Metrics
        if hasattr(self, "_tracker"):
            value = self._tracker.track(value)
        return super().__setattr__(name, value)

    def _check_super_called(self):
        if not hasattr(self, "_tracker"):
            raise RuntimeError(
                "You forgot to call `super().__init__()` "
                "in the `__init__()` method. Go add it!"
            )

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"

    def __str__(self):
        return self.__repr__()



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.language_models.language_model import LanguageModel
from synalinks.src.modules.agents.function_calling_agent import FunctionCallingAgent
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.synthesis.sequential_plan_synthesis import (
    SequentialPlanSynthesis,
)
from synalinks.src.programs.program import Program
from synalinks.src.saving.object_registration import register_synalinks_serializable
from synalinks.src.utils.tool_utils import Tool


@register_synalinks_serializable()
async def calculate(expression: str):
    """Calculate the result of a mathematical expression.

    Args:
        expression (str): The mathematical expression to calculate, such as
            '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
            parentheses, and spaces.
    """
    if not all(char in "0123456789+-*/(). " for char in expression):
        return {
            "result": None,
            "log": "Error: invalid characters in expression",
        }
    try:
        # Evaluate the mathematical expression safely
        result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
        return {
            "result": result,
            "log": "Successfully executed",
        }
    except Exception as e:
        return {
            "result": None,
            "log": f"Error: {e}",
        }


@register_synalinks_serializable()
async def thinking(thinking: str):
    """Think about something.

    Args:
        thinking (str): Your step by step thinking.
    """
    return {
        "thinking": thinking,
    }


class Query(DataModel):
    query: str = Field(
        description="The user query",
    )


class FinalReport(DataModel):
    report: str = Field(
        description="The final report",
    )


class TaskSummary(DataModel):
    summary: str = Field(
        description="The summary of the executed task",
    )


class SequentialPlanSynthesisTest(testing.TestCase):
    async def test_default_synthesis(self):
        language_model = LanguageModel(model="ollama/mistral")

        tools = [
            Tool(calculate),
            Tool(thinking),
        ]

        inputs = Input(data_model=Query)
        outputs = await SequentialPlanSynthesis(
            data_model=FinalReport,
            language_model=language_model,
            runner=FunctionCallingAgent(
                data_model=TaskSummary,
                language_model=language_model,
                tools=tools,
                return_inputs_with_trajectory=False,
            ),
        )(inputs)

        _program = Program(
            inputs=inputs,
            outputs=outputs,
            name="planner_agent",
            description="An agent that learn a step by step plan to achieve a task",
        )

        # Verify program was built successfully
        self.assertIsNotNone(_program)



import heapq
import os
import re
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import duckdb
import duckdb.sqltypes
import orjson

from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend.config import synalinks_home
from synalinks.src.knowledge_bases.database_adapters.database_adapter import (
    DatabaseAdapter,
)
from synalinks.src.utils.async_utils import run_maybe_nested

FTS_KEYS = ["description", "text", "content", "message", "name", "query", "question"]

VSS_KEY = "embedding"

STEMMERS = [
    "arabic",
    "basque",
    "catalan",
    "danish",
    "dutch",
    "english",
    "finnish",
    "french",
    "german",
    "greek",
    "hindi",
    "hungarian",
    "indonesian",
    "irish",
    "italian",
    "lithuanian",
    "nepali",
    "norwegian",
    "porter",
    "portuguese",
    "romanian",
    "russian",
    "serbian",
    "spanish",
    "swedish",
    "tamil",
    "turkish",
    "none",
]

METRICS = [
    "l2seq",
    "cosine",
    "ip",
]

MAIN_TABLE = "main"


def _get_json_columns_from_schema(schema: dict) -> set:
    """Get column names that are JSON type from a JSON schema."""
    json_columns = set()
    properties = schema.get("properties", {})
    for prop_name, prop_spec in properties.items():
        prop_type = prop_spec.get("type")
        if prop_type == "object":
            json_columns.add(prop_name)
        elif prop_type == "array":
            item_spec = prop_spec.get("items", {})
            if item_spec.get("type") == "object":
                json_columns.add(prop_name)
    return json_columns


def _parse_json_columns(row: dict, json_columns: set) -> dict:
    """Parse JSON string columns to Python dicts based on known JSON columns."""
    result = dict(row)
    for col in json_columns:
        if col in result and isinstance(result[col], str):
            try:
                result[col] = orjson.loads(result[col])
            except (orjson.JSONDecodeError, TypeError):
                pass
    return result


def sanitize_identifier(name: str) -> str:
    """Allow only alphanumeric, underscore, and enforce starting with a letter."""
    if not re.match(r"^[A-Za-z_][A-Za-z0-9_]*$", name):
        raise ValueError(f"Invalid SQL identifier: {name}")
    return name


def sanitize_properties(properties: dict):
    """Prevent SQL injections by sanitazing the dict keys for updates.

    Args:
        properties (dict): The properties to sanitize

    Returns:
        (dict): The sanitized properties
    """
    return {sanitize_identifier(k): v for k, v in properties.items()}


class DuckDBAdapter(DatabaseAdapter):
    def __init__(
        self,
        uri=None,
        embedding_model=None,
        data_models=None,
        stemmer="porter",
        metric="cosine",
        fts_keys=FTS_KEYS,
        vss_key=VSS_KEY,
        main_table=MAIN_TABLE,
        vector_dim=None,
        wipe_on_start=False,
        name=None,
        **kwargs,
    ):
        uri = uri.replace("duckdb://", "") if uri else None
        self.uri = uri

        self.embedding_model = embedding_model

        if self.embedding_model:
            if not vector_dim:
                embedded_text = run_maybe_nested(self.embedding_model(["text"]))
                self.vector_dim = len(embedded_text["embeddings"][0])
            else:
                self.vector_dim = vector_dim

        if stemmer not in STEMMERS:
            raise ValueError(f"`stemmer` parameter should be one of {STEMMERS}")
        self.stemmer = stemmer

        if metric not in METRICS:
            raise ValueError(f"`metric` parameter should be one of {METRICS}")
        self.metric = metric

        self.fts_keys = fts_keys
        self.vss_key = vss_key

        self.uri = uri or os.path.join(
            synalinks_home(), name + ".db" if name else "database.db"
        )

        self._install_extensions()

        if wipe_on_start:
            self.wipe_database()

        if data_models:
            self.data_models = data_models
            for dm in data_models:
                self._maybe_create_table(dm)
        else:
            self.data_models = self.get_symbolic_data_models()

    def _connect(self, read_only=False):
        return duckdb.connect(self.uri, read_only=read_only)

    def _install_extensions(self):
        """Install required extensions per-connection."""
        with self._connect(read_only=False) as con:
            con.execute("INSTALL fts;")
            con.execute("LOAD fts;")
            if self.embedding_model:
                con.execute("INSTALL vss;")
                con.execute("LOAD vss;")

    def wipe_database(self):
        with self._connect(read_only=False) as con:
            tables = con.execute(f"""
                SELECT table_name 
                FROM information_schema.tables 
                WHERE table_schema='{MAIN_TABLE}';
            """).fetchall()

            for (table_name,) in tables:
                try:
                    con.execute(f"DROP TABLE IF EXISTS {table_name}")
                except Exception as e:
                    raise RuntimeError(f"Failed to drop table {table_name}: {e}")

    def _get_id_key(self, schema_or_json: dict) -> str:
        """Get the first property key from schema or json as the primary key."""
        if "properties" in schema_or_json:
            props = schema_or_json.get("properties", {})
            if props:
                return next(iter(props.keys()))
        elif schema_or_json:
            return next(iter(schema_or_json.keys()))
        raise ValueError("Cannot determine primary key: schema has no properties")

    def _get_fts_field(self, schema_or_json: dict) -> str | None:
        if "properties" in schema_or_json:
            props = schema_or_json.get("properties")
            for field in self.fts_keys:
                if field in props:
                    return field
            return None
        for field in self.fts_keys:
            if field in schema_or_json:
                return field
        return None

    def _duckdb_table_to_json_schema(
        self,
        table_name: str,
        remove_embedding: bool = True,
    ) -> dict:
        with self._connect(read_only=True) as con:
            info = con.execute(f"PRAGMA table_info('{table_name}')").fetchall()
            props = {}
            for _, name, dtype, _, _, _ in info:
                if name == self.vss_key and remove_embedding:
                    continue
                elif dtype == duckdb.sqltypes.DuckDBPyType(str):
                    props[name] = {"title": name.title(), "type": "string"}
                elif dtype == duckdb.sqltypes.DuckDBPyType(float):
                    props[name] = {"title": name.title(), "type": "number"}
                elif dtype == duckdb.sqltypes.DuckDBPyType(int):
                    props[name] = {"title": name.title(), "type": "integer"}
                elif dtype == duckdb.sqltypes.DuckDBPyType(bool):
                    props[name] = {"title": name.title(), "type": "boolean"}
                elif dtype == duckdb.sqltypes.DuckDBPyType(list[Union[str]]):
                    props[name] = {
                        "title": name.title(),
                        "items": {"type": "string"},
                        "type": "array",
                    }
                elif dtype == duckdb.sqltypes.DuckDBPyType(list[Union[float]]):
                    props[name] = {
                        "title": name.title(),
                        "items": {"type": "number"},
                        "type": "array",
                    }
                elif dtype == duckdb.sqltypes.DuckDBPyType(list[Union[int]]):
                    props[name] = {
                        "title": name.title(),
                        "items": {"type": "integer"},
                        "type": "array",
                    }
                elif dtype == duckdb.sqltypes.DuckDBPyType(list[Union[bool]]):
                    props[name] = {
                        "title": name.title(),
                        "items": {"type": "boolean"},
                        "type": "array",
                    }
                elif dtype == duckdb.sqltypes.DATE:
                    props[name] = {
                        "title": name.title(),
                        "type": "string",
                        "format": "date",
                    }
                elif dtype in (duckdb.sqltypes.TIMESTAMP, duckdb.sqltypes.TIMESTAMP_TZ):
                    props[name] = {
                        "title": name.title(),
                        "type": "string",
                        "format": "date-time",
                    }
                elif dtype == duckdb.sqltypes.TIME:
                    props[name] = {
                        "title": name.title(),
                        "type": "string",
                        "format": "time",
                    }
                elif str(dtype) == "JSON":
                    props[name] = {
                        "title": name.title(),
                        "type": "object",
                    }
                else:
                    raise NotImplementedError(
                        f"Type '{dtype}' not supported by {self.__class__.__name__}"
                        " at the moment, please fill out an issue."
                    )

            return {
                "title": table_name.title(),
                "type": "object",
                "additionalProperties": False,
                "required": list(props.keys()),
                "properties": props,
            }

    def _json_schema_to_duckdb_columns(self, json_schema: dict):
        """Convert JSON schema to DuckDB column definitions.

        Uses the first property as the primary key.
        """
        properties = json_schema.get("properties", {})
        out = []
        first_col = True

        for prop_name, prop_spec in properties.items():
            prop_name = sanitize_identifier(prop_name)
            prop_type = prop_spec.get("type")

            if prop_name == self.vss_key:
                continue

            # Handle anyOf schemas (e.g. Optional[datetime] from Pydantic)
            if not prop_type and "anyOf" in prop_spec:
                for variant in prop_spec["anyOf"]:
                    vtype = variant.get("type")
                    if vtype and vtype != "null":
                        prop_type = vtype
                        prop_spec = variant
                        break

            if not prop_type:
                raise ValueError(
                    f"Malformed JSON schema: "
                    f"missing type for '{prop_name}'"
                )

            col_def = None

            if prop_type == "array":
                item_spec = prop_spec.get("items")
                if not item_spec:
                    # Array without items spec - use JSON
                    col_def = f"{prop_name} JSON"
                else:
                    item_type = item_spec.get("type")
                    if item_type == "string":
                        dtype = duckdb.sqltypes.DuckDBPyType(list[Union[str]])
                        col_def = f"{prop_name} {dtype}"
                    elif item_type == "number":
                        dtype = duckdb.sqltypes.DuckDBPyType(list[Union[float]])
                        col_def = f"{prop_name} {dtype}"
                    elif item_type == "integer":
                        dtype = duckdb.sqltypes.DuckDBPyType(list[Union[int]])
                        col_def = f"{prop_name} {dtype}"
                    elif item_type == "boolean":
                        dtype = duckdb.sqltypes.DuckDBPyType(list[Union[bool]])
                        col_def = f"{prop_name} {dtype}"
                    elif item_type == "object":
                        # Array of objects - use JSON
                        col_def = f"{prop_name} JSON"
                    else:
                        raise ValueError(
                            f"Unsupported array item type '{item_type}' for '{prop_name}'"
                        )
            elif prop_type == "object":
                # Dict/object types stored as JSON
                col_def = f"{prop_name} JSON"
            elif prop_type == "string":
                fmt = prop_spec.get("format")
                if fmt == "date":
                    col_def = f"{prop_name} DATE"
                elif fmt == "date-time":
                    col_def = f"{prop_name} TIMESTAMP"
                elif fmt == "time":
                    col_def = f"{prop_name} TIME"
                else:
                    col_def = f"{prop_name} VARCHAR"
            elif prop_type == "number":
                dtype = duckdb.sqltypes.DuckDBPyType(float)
                col_def = f"{prop_name} {dtype}"
            elif prop_type == "integer":
                dtype = duckdb.sqltypes.DuckDBPyType(int)
                col_def = f"{prop_name} {dtype}"
            elif prop_type == "boolean":
                dtype = duckdb.sqltypes.DuckDBPyType(bool)
                col_def = f"{prop_name} {dtype}"
            else:
                raise ValueError(f"Unsupported JSON schema type: '{prop_type}'")

            # First column becomes primary key
            if first_col and col_def:
                col_def += " PRIMARY KEY"
                first_col = False

            if col_def:
                out.append(col_def)

        if self.embedding_model:
            out.append(f"{self.vss_key} FLOAT[{self.vector_dim}]")
        return ", ".join(out)

    def get_symbolic_data_models(
        self,
        remove_embedding=True,
    ) -> List[SymbolicDataModel]:
        with self._connect(read_only=True) as con:
            tables = con.execute("""
                SELECT table_name 
                FROM information_schema.tables 
                WHERE table_schema='main';
            """).fetchall()

            symbolic_data_models = []
            for (table_name,) in tables:
                schema = self._duckdb_table_to_json_schema(table_name)
                model = SymbolicDataModel(schema=schema, name=table_name)
                symbolic_data_models.append(model)
            return symbolic_data_models

    def _maybe_create_table(
        self,
        data_model: Union[JsonDataModel, SymbolicDataModel],
    ):
        with self._connect(read_only=False) as con:
            json_schema = data_model.get_schema()
            table_name = sanitize_identifier(json_schema.get("title"))

            exists = con.execute(f"""
                SELECT COUNT(*) 
                FROM information_schema.tables 
                WHERE table_schema='{MAIN_TABLE}' AND table_name='{table_name}';
            """).fetchone()[0]

            if exists:
                return

            columns = self._json_schema_to_duckdb_columns(json_schema)
            create_sql = f"CREATE TABLE IF NOT EXISTS {table_name} ({columns});"

            try:
                con.execute(create_sql)
            except Exception as e:
                raise RuntimeError(f"Failed to create table '{table_name}': {e}")

    def _maybe_create_fulltext_index(
        self,
        data_model: Union[JsonDataModel, SymbolicDataModel],
        overwrite: bool = True,
    ):
        with self._connect(read_only=False) as con:
            json_schema = data_model.get_schema()
            table_name = sanitize_identifier(json_schema.get("title"))
            col = self._get_fts_field(schema_or_json=json_schema)
            id_key = self._get_id_key(json_schema)

            if not col:
                return
            con.execute(f"""
                PRAGMA create_fts_index(
                    'main.{table_name}',
                    '{id_key}',
                    '{col}',
                    stemmer='{self.stemmer}',
                    overwrite={1 if overwrite else 0}
                );
            """)

    def _maybe_create_vector_index(
        self,
        data_model: Union[JsonDataModel, SymbolicDataModel],
        overwrite: bool = True,
    ):
        with self._connect(read_only=False) as con:
            json_schema = data_model.get_schema()
            table_name = sanitize_identifier(json_schema.get("title"))
            con.execute(
                f"CREATE INDEX vector_main_{table_name} ON {table_name}"
                f" USING HNSW ({self.vss_key})"
                f" WITH (metric = '{self.metric}');"
            )

    async def update(
        self,
        data_model_or_data_models: Union[List[JsonDataModel], JsonDataModel],
    ) -> Union[Any, List[Any]]:
        """Update or insert records. Returns the primary key value(s)."""
        with self._connect(read_only=False) as con:
            if not isinstance(data_model_or_data_models, list):
                data_models = [data_model_or_data_models]
                return_single = True
            else:
                data_models = data_model_or_data_models
                return_single = False

            ids = []

            for data_model in data_models:
                if not isinstance(data_model, JsonDataModel):
                    data_model = data_model.to_json_data_model()

                self._maybe_create_table(data_model)

                schema = data_model.get_schema()
                table = sanitize_identifier(schema["title"])
                json_data = sanitize_properties(data_model.get_json())
                id_key = self._get_id_key(schema)

                id_val = json_data.get(id_key)
                if id_val is None:
                    raise ValueError(
                        f"Primary key '{id_key}' is required but not provided"
                    )

                cols = list(json_data.keys())
                col_sql = ", ".join(cols)
                placeholders = ", ".join(["?"] * len(cols))
                params = [json_data[c] for c in cols]

                update_cols = [c for c in cols if c != id_key]
                update_sql = ", ".join(f"{c} = EXCLUDED.{c}" for c in update_cols)

                sql = f"""
                    INSERT INTO {table} ({col_sql})
                    VALUES ({placeholders})
                    ON CONFLICT ({id_key}) DO UPDATE SET
                        {update_sql};
                """

                con.execute(sql, params)

                ids.append(id_val)

        self._maybe_create_fulltext_index(data_model)
        return ids[0] if return_single else ids

    async def get(
        self,
        id_or_ids: Any,
        data_models: List[SymbolicDataModel],
        remove_embedding: bool = True,
    ) -> Optional[JsonDataModel]:
        """Get a record by its primary key value."""
        with self._connect(read_only=True) as con:
            if not data_models:
                data_models = self.get_symbolic_data_models()

            for data_model in data_models:
                json_schema = data_model.get_schema()
                table = sanitize_identifier(json_schema.get("title"))
                id_key = self._get_id_key(json_schema)

                try:
                    sql = f"SELECT * FROM {table} WHERE {id_key} = ?"
                    cursor = con.execute(sql, [id_or_ids])
                except Exception as e:
                    warnings.warn(f" {e}")
                    continue

                rows = cursor.arrow().read_all().to_pylist()
                if not rows:
                    continue

                json_columns = _get_json_columns_from_schema(json_schema)
                json_data = _parse_json_columns(rows[0], json_columns)

                if remove_embedding:
                    if self.vss_key in json_data:
                        json_data.pop(self.vss_key)

                id_val = json_data.get(id_key)

                return JsonDataModel(
                    json=json_data,
                    schema=json_schema,
                    name=str(id_val),
                )
            return None

    async def getall(
        self,
        data_model: SymbolicDataModel,
        limit: int = 50,
        offset: int = 0,
        remove_embedding: bool = True,
    ):
        json_schema = data_model.get_schema()
        table = sanitize_identifier(json_schema.get("title"))
        id_key = self._get_id_key(json_schema)

        try:
            with self._connect(read_only=True) as con:
                sql = f"SELECT * FROM {table} LIMIT ? OFFSET ?"
                cursor = con.execute(sql, [limit, offset])
                rows = cursor.arrow().read_all().to_pylist()

                if not rows:
                    return []

                schema = data_model.get_schema()
                json_columns = _get_json_columns_from_schema(schema)

                results = []
                for row in rows:
                    json_data = _parse_json_columns(row, json_columns)

                    if remove_embedding and self.vss_key in json_data:
                        json_data.pop(self.vss_key)

                    results.append(
                        JsonDataModel(
                            json=json_data,
                            schema=schema,
                            name=str(json_data.get(id_key)),
                        )
                    )
                return results
        except Exception as e:
            warnings.warn(f"Failed to read table '{table}': {e}")
            return []

    async def query(self, query: str, params=None, read_only=True, **kwargs):
        with self._connect(read_only=read_only) as con:
            cursor = con.execute(query, params or [])
            return cursor.arrow().read_all().to_pylist()

    async def similarity_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[SymbolicDataModel]] = None,
        k: int = 10,
        threshold: float = None,
    ):
        with self._connect(read_only=True) as con:
            if not text_or_texts:
                return []

            texts = (
                [text_or_texts] if not isinstance(text_or_texts, list) else text_or_texts
            )

            if not data_models:
                data_models = self.get_symbolic_data_models()

            results = {}

            for model in data_models:
                schema = model.get_schema()
                label = sanitize_identifier(schema.get("title"))
                id_key = self._get_id_key(schema)

                vectors = await self.embedding_model(texts)
                vectors = vectors.get("embeddings")

                for vector in vectors:
                    where_clause = (
                        (
                            f"WHERE array_distance({self.vss_key}, "
                            f"?::FLOAT[{self.vector_dim}]) < ?"
                        )
                        if threshold
                        else ""
                    )

                    sql = f"""
                        SELECT *,
                            array_distance(
                                {self.vss_key}, ?::FLOAT[{self.vector_dim}]
                            ) AS score
                        FROM {label}
                        {where_clause}
                        ORDER BY score ASC
                        LIMIT ?;
                    """

                    params = [vector]
                    if threshold is not None:
                        params.extend([vector, threshold])
                    params.append(k)

                    try:
                        cursor = con.execute(sql, params)
                        rows = cursor.arrow().read_all().to_pylist()
                    except Exception as e:
                        raise RuntimeError(
                            f"Vector search failed for table '{label}': {e}"
                        )

                    for row in rows:
                        results[row[id_key]] = row

            ranked = sorted(results.values(), key=lambda r: r["score"])
            return ranked[:k]

    async def fulltext_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[SymbolicDataModel]] = None,
        k: int = 10,
        threshold: float = None,
    ):
        if not text_or_texts:
            return []

        texts = [text_or_texts] if not isinstance(text_or_texts, list) else text_or_texts

        if not data_models:
            data_models = self.get_symbolic_data_models()

        results = {}
        with self._connect(read_only=True) as con:
            for model in data_models:
                schema = model.get_schema()
                label = sanitize_identifier(schema.get("title"))
                id_key = self._get_id_key(schema)
                col = self._get_fts_field(schema)
                if not col:
                    warnings.warn(f"Skipping FTS search for {label}: no FTS field found.")
                    continue

                fts_table = sanitize_identifier(f"fts_main_{label}")

                for text in texts:
                    sql = f"""
                        SELECT t.*, fts.score
                        FROM {label} t
                        JOIN (
                            SELECT
                                {id_key},
                                {fts_table}.match_bm25({id_key}, ?) AS score
                            FROM {label}
                        ) fts ON t.{id_key} = fts.{id_key}
                        WHERE fts.score IS NOT NULL
                        ORDER BY fts.score DESC
                        LIMIT ?;
                    """
                    params = [text]
                    if threshold:
                        params.append(threshold)
                    params.append(k)
                    try:
                        cursor = con.execute(sql, params)
                        rows = cursor.arrow().read_all().to_pylist()
                    except Exception as e:
                        raise RuntimeError(f"FTS query failed for table '{label}': {e}")

                    for row in rows:
                        results[row[id_key]] = row

            ranked = sorted(results.values(), key=lambda r: r["score"])
            return ranked[:k]

    async def hybrid_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[SymbolicDataModel]] = None,
        k: int = 10,
        k_rank: int = 60,
        similarity_threshold: float = None,
        fulltext_threshold: float = None,
    ):
        """
        Perform hybrid search using Reciprocal Rank Fusion (RRF).

        Combines ranks from:
          - Full-text search (BM25-based)
          - Vector similarity search (embedding distance)

        Args:
            text_or_texts: Single query or list of queries.
            data_models: Symbolic data models to search (optional).
            k: Maximum number of results to return.
            similarity_threshold: Optional maximum distance for vector search
                (lower = better).
            fulltext_threshold: Optional minimum BM25 score for text search
                (higher = better).
            k_rank: RRF smoothing constant. Lower emphasizes top ranks (default=60).

        Returns:
            List of dicts containing merged, reranked results with "score".
        """

        if not text_or_texts:
            return []

        if not self.embedding_model:
            return await self.fulltext_search(text_or_texts, data_models, k)

        queries = [text_or_texts] if isinstance(text_or_texts, str) else text_or_texts
        if not data_models:
            data_models = self.get_symbolic_data_models()

        final_results = {}

        for query_text in queries:
            try:
                try:
                    fts_results = await self.fulltext_search(
                        query_text, data_models, k=k * 5, threshold=fulltext_threshold
                    )
                except Exception:
                    fts_results = []

                try:
                    vss_results = await self.similarity_search(
                        query_text, data_models, k=k * 5, threshold=similarity_threshold
                    )
                except Exception:
                    vss_results = []

                if not fts_results and not vss_results:
                    warnings.warn(f"No results for query='{query_text}'.")
                    continue

                # Get id_key from first data model
                id_key = self._get_id_key(data_models[0].get_schema())

                fts_rank = {r[id_key]: i + 1 for i, r in enumerate(fts_results)}
                vss_rank = {r[id_key]: i + 1 for i, r in enumerate(vss_results)}

                combined_rows: Dict[str, Dict[str, Any]] = {}

                for row in fts_results + vss_results:
                    uid = row[id_key]
                    if uid not in combined_rows:
                        combined_rows[uid] = dict(row)
                    else:
                        combined_rows[uid].update(row)

                # RRF formula: sum(1 / (k_rank + rank))
                for uid in set(fts_rank) | set(vss_rank):
                    score = 0.0
                    if uid in fts_rank:
                        score += 1.0 / (k_rank + fts_rank[uid])
                    if uid in vss_rank:
                        score += 1.0 / (k_rank + vss_rank[uid])
                    combined_rows[uid]["score"] = score

                top_rows = heapq.nlargest(
                    k, combined_rows.values(), key=lambda r: r["score"]
                )

                for r in top_rows:
                    uid = r[id_key]
                    if (
                        uid not in final_results
                        or r["score"] > final_results[uid]["score"]
                    ):
                        final_results[uid] = r

            except Exception as e:
                warnings.warn(f"Hybrid search iteration failed: {e}")
                continue

        results_sorted = heapq.nlargest(
            k, final_results.values(), key=lambda r: r["score"]
        )

        # Get id_key for final sorting
        if data_models:
            id_key = self._get_id_key(data_models[0].get_schema())
        else:
            id_key = next(iter(results_sorted[0].keys())) if results_sorted else "id"

        return sorted(results_sorted, key=lambda r: (-r["score"], r.get(id_key)))



# Modified from: keras/src/trainers/data_adapters/array_data_adapter_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from absl.testing import parameterized

from synalinks.src import testing
from synalinks.src.testing.test_utils import AnswerWithRationale
from synalinks.src.testing.test_utils import Query
from synalinks.src.testing.test_utils import load_test_data
from synalinks.src.testing.test_utils import named_product
from synalinks.src.trainers.data_adapters.array_data_adapter import ArrayDataAdapter


class TestArrayDataAdapter(testing.TestCase):
    @parameterized.named_parameters(
        named_product(
            shuffle=[False, "batch", True],
        )
    )
    def test_basic_flow(self, shuffle):
        (x, y), (x_test, y_test) = load_test_data()

        adapter = ArrayDataAdapter(
            x,
            y=y,
            batch_size=32,
            steps=None,
            shuffle=shuffle,
        )
        self.assertEqual(adapter.num_batches, 1)
        self.assertEqual(adapter.batch_size, 32)
        self.assertEqual(adapter.has_partial_batch, True)
        self.assertEqual(adapter.partial_batch_size, 15)

        it = adapter.get_numpy_iterator()
        for i, batch in enumerate(it):
            self.assertEqual(len(batch), 2)
            x, y = batch
            self.assertIsInstance(batch, tuple)
            self.assertIsInstance(x[0], Query)
            self.assertIsInstance(x[1], Query)
            self.assertIsInstance(y[0], AnswerWithRationale)
            self.assertIsInstance(y[1], AnswerWithRationale)



from synalinks.src.knowledge_bases.database_adapters.database_adapter import (
    DatabaseAdapter,
)
from synalinks.src.knowledge_bases.database_adapters.duckdb_adapter import DuckDBAdapter


def get(uri):
    if not uri:
        return DuckDBAdapter
    if uri.startswith("duckdb"):
        return DuckDBAdapter
    else:
        return DuckDBAdapter



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import is_symbolic_data_model
from synalinks.src.knowledge_bases import database_adapters
from synalinks.src.saving import serialization_lib
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable
from synalinks.src.utils.naming import auto_name


@synalinks_export("synalinks.KnowledgeBase")
class KnowledgeBase(SynalinksSaveable):
    """A knowledge base for storing and retrieving structured data.

    The KnowledgeBase provides a unified interface for storing structured data
    with support for full-text search and optional vector similarity search.
    It uses DuckDB as the underlying storage engine.

    ### Basic Usage

    ```python
    import synalinks

    class Document(synalinks.DataModel):
        id: str
        title: str
        content: str

    # Create a knowledge base without embeddings (full-text search only)
    knowledge_base = synalinks.KnowledgeBase(
        uri="duckdb://my_database.db",
        data_models=[Document],
    )

    # Store a document
    doc = Document(id="1", title="Hello", content="Hello World!")
    await knowledge_base.update(doc.to_json_data_model())

    # Retrieve by ID
    result = await knowledge_base.get("1", [Document.to_symbolic_data_model()])

    # Full-text search
    results = await knowledge_base.fulltext_search("Hello", k=10)
    ```

    ### With Vector Similarity Search

    ```python
    embedding_model = synalinks.EmbeddingModel(
        model="ollama/mxbai-embed-large"
    )

    knowledge_base = synalinks.KnowledgeBase(
        uri="duckdb://./my_database.db",
        data_models=[Document],
        embedding_model=embedding_model,
        metric="cosine",
    )

    # Hybrid search (combines full-text and vector similarity)
    results = await knowledge_base.hybrid_search("semantic query", k=10)
    ```

    ### Retrieving Table Definitions

    ```python
    # Get all symbolic data models (table definitions) from the database
    symbolic_models = knowledge_base.get_symbolic_data_models()

    for model in symbolic_models:
        print(model.get_schema())
        # {'title': 'Document', 'type': 'object', 'properties': {...}, ...}
    ```

    Args:
        uri (str): The database connection URI. Use "duckdb://path/to/db.db"
            for DuckDB. If not provided, uses an in-memory database.
        data_models (list): Optional list of DataModel or SymbolicDataModel
            classes to create tables for.
        embedding_model (EmbeddingModel): Optional embedding model for
            vector similarity search.
        metric (str): The distance metric for vector search.
            Options: "cosine", "l2seq", "ip" (default: "cosine").
        wipe_on_start (bool): Whether to clear the database on initialization
            (default: False).
        name (str): Optional name for the knowledge base (used for serialization).
    """

    def __init__(
        self,
        uri=None,
        data_models=None,
        embedding_model=None,
        metric="cosine",
        wipe_on_start=False,
        name=None,
    ):
        self.adapter = database_adapters.get(uri)(
            uri=uri,
            data_models=data_models,
            embedding_model=embedding_model,
            metric=metric,
            wipe_on_start=wipe_on_start,
            name=name,
        )
        self.uri = uri
        self.data_models = data_models or []
        self.embedding_model = embedding_model
        self.metric = metric
        self.wipe_on_start = wipe_on_start
        if not name:
            self.name = auto_name("knowledge_base")
        else:
            self.name = name

    async def update(
        self,
        data_model_or_data_models: Union[Any, List[Any]],
    ) -> Union[Any, List[Any]]:
        """Insert or update records in the knowledge base.

        Args:
            data_model_or_data_models (JsonDataModel | List[JsonDataModel]):
                A single JsonDataModel or a list of JsonDataModels to insert
                or update. Uses the first field as the primary key for upserts.

        Returns:
            The primary key value(s) of the inserted/updated records.
        """
        return await self.adapter.update(data_model_or_data_models)

    async def get(
        self,
        id_or_ids: Any,
        data_models: Optional[List[Any]] = None,
    ) -> Optional[Any]:
        """Retrieve a record by its primary key.

        Args:
            id_or_ids: The primary key value to look up.
            data_models: Optional list of SymbolicDataModels to search in.
                If not provided, searches all tables.

        Returns:
            JsonDataModel if found, None otherwise.
        """
        return await self.adapter.get(id_or_ids, data_models=data_models)

    async def getall(
        self,
        data_model: Any,
        limit: int = 50,
        offset: int = 0,
    ) -> List[Any]:
        """Retrieve all records from a table with pagination.

        Args:
            data_model: The SymbolicDataModel representing the table to query.
            limit: Maximum number of records to return (default: 50).
            offset: Number of records to skip (default: 0).

        Returns:
            List of JsonDataModels.
        """
        return await self.adapter.getall(data_model, limit=limit, offset=offset)

    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """Execute a raw SQL query against the knowledge base.

        Args:
            query (str): The SQL query to execute.
            params (dict): Optional list of parameters for parameterized queries.
            **kwargs (Any): Additional options (e.g., read_only=True/False).

        Returns:
            List of result dictionaries.
        """
        return await self.adapter.query(query, params=params, **kwargs)

    async def similarity_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[Any]] = None,
        k: int = 10,
        threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Perform vector similarity search using embeddings.

        Requires an embedding_model to be configured.

        Args:
            text_or_texts: Query text or list of query texts.
            data_models: Optional list of SymbolicDataModels to search in.
            k: Maximum number of results to return (default: 10).
            threshold: Optional maximum distance threshold for filtering.

        Returns:
            List of matching records with similarity scores.
        """
        return await self.adapter.similarity_search(
            text_or_texts,
            data_models=data_models,
            k=k,
            threshold=threshold,
        )

    async def fulltext_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[Any]] = None,
        k: int = 10,
        threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Perform full-text search using BM25 ranking.

        Searches text fields (description, text, content, message, name,
        query, question) for matching documents.

        Args:
            text_or_texts: Query text or list of query texts.
            data_models: Optional list of SymbolicDataModels to search in.
            k: Maximum number of results to return (default: 10).
            threshold: Optional minimum BM25 score threshold.

        Returns:
            List of matching records with relevance scores.
        """
        return await self.adapter.fulltext_search(
            text_or_texts,
            data_models=data_models,
            k=k,
            threshold=threshold,
        )

    async def hybrid_search(
        self,
        text_or_texts: Union[str, List[str]],
        data_models: Optional[List[Any]] = None,
        k: int = 10,
        k_rank: int = 60,
        similarity_threshold: Optional[float] = None,
        fulltext_threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Perform hybrid search combining vector similarity and full-text.

        Uses Reciprocal Rank Fusion (RRF) to combine results from both
        similarity search and full-text search. Falls back to full-text
        search only if no embedding model is configured.

        Args:
            text_or_texts: Query text or list of query texts.
            data_models: Optional list of SymbolicDataModels to search in.
            k: Maximum number of results to return (default: 10).
            k_rank: RRF smoothing constant. Lower values emphasize top ranks
                more strongly (default: 60).
            similarity_threshold: Optional threshold for vector similarity.
            fulltext_threshold: Optional threshold for full-text relevance.

        Returns:
            List of matching records with combined scores.
        """
        return await self.adapter.hybrid_search(
            text_or_texts,
            data_models=data_models,
            k=k,
            k_rank=k_rank,
            similarity_threshold=similarity_threshold,
            fulltext_threshold=fulltext_threshold,
        )

    def get_symbolic_data_models(self) -> List[Any]:
        """Retrieve all symbolic data models (table definitions) from the database.

        Returns a list of SymbolicDataModel objects representing each table
        in the database. This is useful for introspecting the database schema
        or for passing to search methods to limit the search scope.

        Returns:
            list: List of symbolic data models representing the database tables.

        Example:
            ```python
            symbolic_models = knowledge_base.get_symbolic_data_models()
            for model in symbolic_models:
                schema = model.get_schema()
                print(f"Table: {schema['title']}")
                print(f"Fields: {list(schema['properties'].keys())}")
            ```
        """
        return self.adapter.get_symbolic_data_models()

    def get_config(self):
        config = {
            "uri": self.uri,
            "name": self.name,
            "metric": self.metric,
            "wipe_on_start": self.wipe_on_start,
        }
        data_models_config = {
            "data_models": [
                (
                    serialization_lib.serialize_synalinks_object(
                        data_model.to_symbolic_data_model(
                            name="data_model" + (f"_{i}_" if i > 0 else "_") + self.name
                        )
                    )
                    if not is_symbolic_data_model(data_model)
                    else serialization_lib.serialize_synalinks_object(data_model)
                )
                for i, data_model in enumerate(self.data_models)
            ]
        }
        embedding_model_config = {}
        if self.embedding_model:
            embedding_model_config = {
                "embedding_model": serialization_lib.serialize_synalinks_object(
                    self.embedding_model,
                )
            }
        return {
            **data_models_config,
            **embedding_model_config,
            **config,
        }

    @classmethod
    def from_config(cls, config):
        data_models_config = config.pop("data_models", [])
        data_models = [
            serialization_lib.deserialize_synalinks_object(data_model)
            for data_model in data_models_config
        ]
        embedding_model = None
        if "embedding_model" in config:
            embedding_model = serialization_lib.deserialize_synalinks_object(
                config.pop("embedding_model"),
            )
        return cls(
            data_models=data_models,
            embedding_model=embedding_model,
            **config,
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.modules.core.identity import Identity
from synalinks.src.modules.core.input_module import Input
from synalinks.src.programs.program import Program


class IdentityTest(testing.TestCase):
    async def test_single_inputs(self):
        class Query(DataModel):
            query: str

        inputs = Input(data_model=Query)
        outputs = await Identity()(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        result = await program(Query(query="a"))

        expected_json = {
            "query": "a",
        }

        self.assertEqual(result.get_json(), expected_json)

    async def test_tuple_inputs(self):
        class Query(DataModel):
            query: str

        inputs = (
            Input(data_model=Query),
            Input(data_model=Query),
            Input(data_model=Query),
        )
        outputs = await Identity()(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        result = await program(
            (
                Query(query="a"),
                Query(query="b"),
                Query(query="c"),
            )
        )

        expected_json_a = {
            "query": "a",
        }
        expected_json_b = {
            "query": "b",
        }
        expected_json_c = {
            "query": "c",
        }

        self.assertEqual(result[0].get_json(), expected_json_a)
        self.assertEqual(result[1].get_json(), expected_json_b)
        self.assertEqual(result[2].get_json(), expected_json_c)

    async def test_list_inputs(self):
        class Query(DataModel):
            query: str

        inputs = [
            Input(data_model=Query),
            Input(data_model=Query),
            Input(data_model=Query),
        ]
        outputs = await Identity()(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        result = await program(
            [
                Query(query="a"),
                Query(query="b"),
                Query(query="c"),
            ]
        )

        expected_json_a = {
            "query": "a",
        }
        expected_json_b = {
            "query": "b",
        }
        expected_json_c = {
            "query": "c",
        }

        self.assertEqual(result[0].get_json(), expected_json_a)
        self.assertEqual(result[1].get_json(), expected_json_b)
        self.assertEqual(result[2].get_json(), expected_json_c)

    async def test_dict_inputs(self):
        class Query(DataModel):
            query: str

        inputs = {
            "a": Input(data_model=Query),
            "b": Input(data_model=Query),
            "c": Input(data_model=Query),
        }
        outputs = await Identity()(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        result = await program(
            {
                "a": Query(query="a"),
                "b": Query(query="b"),
                "c": Query(query="c"),
            }
        )

        expected_json_a = {
            "query": "a",
        }
        expected_json_b = {
            "query": "b",
        }
        expected_json_c = {
            "query": "c",
        }

        self.assertEqual(result["a"].get_json(), expected_json_a)
        self.assertEqual(result["b"].get_json(), expected_json_b)
        self.assertEqual(result["c"].get_json(), expected_json_c)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import numpy as np

from synalinks.src.backend import DataModel


class Query(DataModel):
    query: str


class AnswerWithRationale(DataModel):
    rationale: str
    answer: str


def load_test_data():
    x_train = np.array(
        [
            Query(query="What is the capital of France?"),
            Query(query="What is the French city of aeronautics?"),
            Query(query="Which city is known as the fashion capital of the world?"),
            Query(query="What is the highest mountain in France?"),
            Query(query="Which river flows through Paris?"),
            Query(query="What is the capital of Italy?"),
            Query(query="Which mountain range separates France and Spain?"),
            Query(query="What is the largest lake in France?"),
            Query(query="Which French city is famous for its wine production?"),
            Query(query="What is the currency used in France?"),
            Query(query="Which ocean borders western France?"),
            Query(query="What is the official language of France?"),
            Query(query="Which French region is known for lavender fields?"),
            Query(query="What is the most visited monument in Paris?"),
            Query(query="Which French city hosted the 1968 Winter Olympics?"),
        ],
        dtype="object",
    )
    y_train = np.array(
        [
            AnswerWithRationale(
                rationale="""The capital of France is well-known and is the seat of """
                """the French government.""",
                answer="Paris",
            ),
            AnswerWithRationale(
                rationale="""Toulouse is known as the French city of aeronautics due"""  # noqa: E501
                """ to its significant contributions to the aerospace industry.""",
                answer="Toulouse",
            ),
            AnswerWithRationale(
                rationale="""Paris is widely recognized as the fashion capital of the world, """  # noqa: E501
                """hosting major fashion weeks and luxury brands.""",
                answer="Paris",
            ),
            AnswerWithRationale(
                rationale="""Mont Blanc is the highest mountain in France, standing at """  # noqa: E501
                """4,807 meters above sea level.""",
                answer="Mont Blanc",
            ),
            AnswerWithRationale(
                rationale="""The Seine River flows through Paris, dividing the city """
                """and serving as a major waterway.""",
                answer="Seine",
            ),
            AnswerWithRationale(
                rationale="""Rome is the capital city of Italy and has been the center """  # noqa: E501
                """of Italian government and culture for centuries.""",
                answer="Rome",
            ),
            AnswerWithRationale(
                rationale="""The Pyrenees mountain range forms a natural border """
                """between France and Spain.""",
                answer="Pyrenees",
            ),
            AnswerWithRationale(
                rationale="""Lake Geneva (Lac Léman) is the largest lake in France, """  # noqa: E501
                """shared with Switzerland.""",
                answer="Lake Geneva",
            ),
            AnswerWithRationale(
                rationale="""Bordeaux is renowned worldwide for its wine production """  # noqa: E501
                """and vineyard regions.""",
                answer="Bordeaux",
            ),
            AnswerWithRationale(
                rationale="""France uses the Euro as its official currency, """
                """adopted in 1999 as part of the European Union.""",
                answer="Euro",
            ),
            AnswerWithRationale(
                rationale="""The Atlantic Ocean borders the western coast of France, """  # noqa: E501
                """providing important maritime access.""",
                answer="Atlantic Ocean",
            ),
            AnswerWithRationale(
                rationale="""French is the official language of France, spoken """
                """by the vast majority of the population.""",
                answer="French",
            ),
            AnswerWithRationale(
                rationale="""Provence is famous for its extensive lavender fields, """
                """particularly around the Valensole plateau.""",
                answer="Provence",
            ),
            AnswerWithRationale(
                rationale="""The Eiffel Tower is the most visited paid monument """
                """in the world and Paris's most iconic landmark.""",
                answer="Eiffel Tower",
            ),
            AnswerWithRationale(
                rationale="""Grenoble hosted the 1968 Winter Olympics, becoming """
                """an important winter sports destination.""",
                answer="Grenoble",
            ),
        ],
        dtype="object",
    )

    x_test = np.array(
        [
            Query(query="What is the largest city in France?"),  # noqa: E501
            Query(query="What is the capital of Germany?"),  # noqa: E501
            Query(query="Which French city is known for its film festival?"),  # noqa: E501
            Query(query="What is the longest river in France?"),  # noqa: E501
            Query(query="Which French region is famous for champagne?"),  # noqa: E501
            Query(query="What is the capital of Spain?"),  # noqa: E501
            Query(query="Which sea borders southern France?"),  # noqa: E501
            Query(query="What is the national symbol of France?"),  # noqa: E501
            Query(query="Which French city is home to the European Parliament?"),  # noqa: E501
            Query(query="What is the highest point in the French Alps?"),  # noqa: E501
        ],
        dtype="object",
    )
    y_test = np.array(
        [
            AnswerWithRationale(
                rationale="Paris is the largest city in France by population and area.",  # noqa: E501
                answer="Paris",
            ),
            AnswerWithRationale(
                rationale="The capital of Germany is well-known and is the seat of the German government.",  # noqa: E501
                answer="Berlin",
            ),
            AnswerWithRationale(
                rationale="Cannes is famous for hosting the prestigious Cannes Film Festival annually.",  # noqa: E501
                answer="Cannes",
            ),
            AnswerWithRationale(
                rationale="The Loire River is the longest river in France, flowing for 1,012 kilometers.",  # noqa: E501
                answer="Loire",
            ),
            AnswerWithRationale(
                rationale="Champagne region is world-famous for producing the sparkling wine of the same name.",  # noqa: E501
                answer="Champagne",
            ),
            AnswerWithRationale(
                rationale="Madrid is the capital and largest city of Spain, located in the center of the country.",  # noqa: E501
                answer="Madrid",
            ),
            AnswerWithRationale(
                rationale="The Mediterranean Sea borders the southern coast of France along the French Riviera.",  # noqa: E501
                answer="Mediterranean Sea",
            ),
            AnswerWithRationale(
                rationale="The rooster (coq) is the national symbol of France, representing vigilance and pride.",  # noqa: E501
                answer="Rooster",
            ),
            AnswerWithRationale(
                rationale="Strasbourg is home to the European Parliament and serves as a major EU political center.",  # noqa: E501
                answer="Strasbourg",
            ),
            AnswerWithRationale(
                rationale="Mont Blanc is the highest point in the French Alps and in Western Europe at 4,807 meters.",  # noqa: E501
                answer="Mont Blanc",
            ),
        ],
        dtype="object",
    )
    return (x_train, y_train), (x_test, y_test)


def mock_completion_data():
    # Training responses - expanded to match new training data
    responses = [
        """{"rationale":"The capital of France is well-known and is the seat of the French government.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"Toulouse is known as the French city of aeronautics due to its significant contributions to the aerospace industry.", "answer": "Toulouse"}""",  # noqa: E501
        """{"rationale":"Paris is widely recognized as the fashion capital of the world, hosting major fashion weeks and luxury brands.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"Mont Blanc is the highest mountain in France, standing at 4,807 meters above sea level.", "answer": "Mont Blanc"}""",  # noqa: E501
        """{"rationale":"The Seine River flows through Paris, dividing the city and serving as a major waterway.", "answer": "Seine"}""",  # noqa: E501
        """{"rationale":"Rome is the capital city of Italy and has been the center of Italian government and culture for centuries.", "answer": "Rome"}""",  # noqa: E501
        """{"rationale":"The Pyrenees mountain range forms a natural border between France and Spain.", "answer": "Pyrenees"}""",  # noqa: E501
        """{"rationale":"Lake Geneva (Lac Léman) is the largest lake in France, shared with Switzerland.", "answer": "Lake Geneva"}""",  # noqa: E501
        """{"rationale":"Bordeaux is renowned worldwide for its wine production and vineyard regions.", "answer": "Bordeaux"}""",  # noqa: E501
        """{"rationale":"France uses the Euro as its official currency, adopted in 1999 as part of the European Union.", "answer": "Euro"}""",  # noqa: E501
        """{"rationale":"The Atlantic Ocean borders the western coast of France, providing important maritime access.", "answer": "Atlantic Ocean"}""",  # noqa: E501
        """{"rationale":"French is the official language of France, spoken by the vast majority of the population.", "answer": "French"}""",  # noqa: E501
        """{"rationale":"Provence is famous for its extensive lavender fields, particularly around the Valensole plateau.", "answer": "Provence"}""",  # noqa: E501
        """{"rationale":"The Eiffel Tower is the most visited paid monument in the world and Paris's most iconic landmark.", "answer": "Eiffel Tower"}""",  # noqa: E501
        """{"rationale":"Grenoble hosted the 1968 Winter Olympics, becoming an important winter sports destination.", "answer": "Grenoble"}""",  # noqa: E501
    ]
    return [{"choices": [{"message": {"content": response}}]} for response in responses]


def mock_incorrect_completion_data():
    # Training responses with some incorrect answers
    responses = [
        """{"rationale":"The capital of France is well-known and is the seat of the French government.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"Paris is known as the French city of aeronautics due to its significant contributions to the aerospace industry.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"Milan is widely recognized as the fashion capital of the world, hosting major fashion weeks.", "answer": "Milan"}""",  # noqa: E501
        """{"rationale":"Mount Everest is the highest mountain in France, standing at great height.", "answer": "Mount Everest"}""",  # noqa: E501
        """{"rationale":"The Thames River flows through Paris, dividing the city.", "answer": "Thames"}""",  # noqa: E501
        """{"rationale":"Paris is the capital city of Italy and has been the center of Italian government.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"The Alps mountain range forms a natural border between France and Spain.", "answer": "Alps"}""",  # noqa: E501
        """{"rationale":"Lake Superior is the largest lake in France, shared with other countries.", "answer": "Lake Superior"}""",  # noqa: E501
        """{"rationale":"Paris is renowned worldwide for its wine production and vineyard regions.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"France uses the Dollar as its official currency in modern times.", "answer": "Dollar"}""",  # noqa: E501
        """{"rationale":"The Pacific Ocean borders the western coast of France.", "answer": "Pacific Ocean"}""",  # noqa: E501
        """{"rationale":"English is the official language of France, spoken by most people.", "answer": "English"}""",  # noqa: E501
        """{"rationale":"Paris is famous for its extensive lavender fields in the countryside.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"The Statue of Liberty is the most visited monument in Paris.", "answer": "Statue of Liberty"}""",  # noqa: E501
        """{"rationale":"Paris hosted the 1968 Winter Olympics, becoming a winter sports destination.", "answer": "Paris"}""",  # noqa: E501
    ]
    return [{"choices": [{"message": {"content": response}}]} for response in responses]


def mock_test_completion_data():
    # Test responses - expanded to match new test data
    responses = [
        """{"rationale":"Paris is the largest city in France by population and area.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"The capital of Germany is well-known and is the seat of the German government.", "answer": "Berlin"}""",  # noqa: E501
        """{"rationale":"Cannes is famous for hosting the prestigious Cannes Film Festival annually.", "answer": "Cannes"}""",  # noqa: E501
        """{"rationale":"The Loire River is the longest river in France, flowing for 1,012 kilometers.", "answer": "Loire"}""",  # noqa: E501
        """{"rationale":"Champagne region is world-famous for producing the sparkling wine of the same name.", "answer": "Champagne"}""",  # noqa: E501
        """{"rationale":"Madrid is the capital and largest city of Spain, located in the center of the country.", "answer": "Madrid"}""",  # noqa: E501
        """{"rationale":"The Mediterranean Sea borders the southern coast of France along the French Riviera.", "answer": "Mediterranean Sea"}""",  # noqa: E501
        """{"rationale":"The rooster (coq) is the national symbol of France, representing vigilance and pride.", "answer": "Rooster"}""",  # noqa: E501
        """{"rationale":"Strasbourg is home to the European Parliament and serves as a major EU political center.", "answer": "Strasbourg"}""",  # noqa: E501
        """{"rationale":"Mont Blanc is the highest point in the French Alps and in Western Europe at 4,807 meters.", "answer": "Mont Blanc"}""",  # noqa: E501
    ]
    return [{"choices": [{"message": {"content": response}}]} for response in responses]


def mock_incorrect_test_completion_data():
    # Incorrect test responses, wrong answers for x_test queries
    responses = [
        """{"rationale":"Lyon is the largest city in France by population and area.", "answer": "Lyon"}""",  # noqa: E501
        """{"rationale":"The capital of Germany is Munich, which is a major city in southern Germany.", "answer": "Munich"}""",  # noqa: E501
        """{"rationale":"Paris is famous for hosting the prestigious film festival annually.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"The Seine River is the longest river in France, flowing through major cities.", "answer": "Seine"}""",  # noqa: E501
        """{"rationale":"Bordeaux region is world-famous for producing the sparkling wine.", "answer": "Bordeaux"}""",  # noqa: E501
        """{"rationale":"Barcelona is the capital and largest city of Spain, located on the coast.", "answer": "Barcelona"}""",  # noqa: E501
        """{"rationale":"The Atlantic Ocean borders the southern coast of France.", "answer": "Atlantic Ocean"}""",  # noqa: E501
        """{"rationale":"The eagle is the national symbol of France, representing strength and freedom.", "answer": "Eagle"}""",  # noqa: E501
        """{"rationale":"Paris is home to the European Parliament and serves as a major EU center.", "answer": "Paris"}""",  # noqa: E501
        """{"rationale":"The Eiffel Tower is the highest point in the French Alps.", "answer": "Eiffel Tower"}""",  # noqa: E501
    ]
    return [{"choices": [{"message": {"content": response}}]} for response in responses]


def named_product(*args, **kwargs):
    """Utility to generate the cartesian product of parameters values and
    generate a test case names for each combination.

    The result of this function is to be used with the
    `@parameterized.named_parameters` decorator. It is a replacement for
    `@parameterized.product` which adds explicit test case names.

    For example, this code:
    ```
    class NamedExample(parameterized.TestCase):
        @parameterized.named_parameters(
            named_product(
                [
                    {'testcase_name': 'negative', 'x': -1},
                    {'testcase_name': 'positive', 'x': 1},
                    {'testcase_name': 'zero', 'x': 0},
                ],
                numeral_type=[float, int],
            )
        )
        def test_conversion(self, x, numeral_type):
            self.assertEqual(numeral_type(x), x)
    ```
    produces six tests (note that absl will reorder them by name):
    - `NamedExample::test_conversion_negative_float`
    - `NamedExample::test_conversion_positive_float`
    - `NamedExample::test_conversion_zero_float`
    - `NamedExample::test_conversion_negative_int`
    - `NamedExample::test_conversion_positive_int`
    - `NamedExample::test_conversion_zero_int`

    This function is also useful in the case where there is no product to
    generate test case names for one argument:
    ```
    @parameterized.named_parameters(named_product(numeral_type=[float, int]))
    ```

    Args:
        *args: Each positional parameter is a sequence of keyword arg dicts.
            Every test case generated will include exactly one dict from each
            positional parameter. These will then be merged to form an overall
            list of arguments for the test case. Each dict must contain a
            `"testcase_name"` key whose value is combined with others to
            generate the test case name.
        **kwargs: A mapping of parameter names and their possible values.
            Possible values should given as either a list or a tuple. A string
            representation of each value is used to generate the test case name.

    Returns:
        A list of maps for the test parameters combinations to pass to
        `@parameterized.named_parameters`.
    """

    def value_to_str(value):
        if hasattr(value, "__name__"):
            return value.__name__.lower()
        return str(value).lower()

    # Convert the keyword arguments in the same dict format as the args
    all_test_dicts = args + tuple(
        tuple({"testcase_name": value_to_str(v), key: v} for v in values)
        for key, values in kwargs.items()
    )

    # The current list of tests, start with one empty test
    tests = [{}]
    for test_dicts in all_test_dicts:
        new_tests = []
        for test_dict in test_dicts:
            for test in tests:
                # Augment the testcase name by appending
                testcase_name = test.get("testcase_name", "")
                testcase_name += "_" if testcase_name else ""
                testcase_name += test_dict["testcase_name"]
                new_test = test.copy()
                # Augment the test by adding all the parameters
                new_test.update(test_dict)
                new_test["testcase_name"] = testcase_name
                new_tests.append(new_test)
        # Overwrite the list of tests with the product obtained so far
        tests = new_tests

    return tests



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import warnings

import litellm

from synalinks.src.api_export import synalinks_export
from synalinks.src.saving import serialization_lib
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable

litellm.disable_aiohttp_transport = True


@synalinks_export(
    [
        "synalinks.EmbeddingModel",
        "synalinks.embedding_models.EmbeddingModel",
    ]
)
class EmbeddingModel(SynalinksSaveable):
    """An embedding model API wrapper.

    Embedding models are a type of machine learning model used to convert
    high-dimensional data, such as text into lower-dimensional vector
    representations while preserving the semantic meaning and relationships
    within the data. These vector representations, known as embeddings,
    allow for more efficient and effective processing in various tasks.

    Many providers are available like OpenAI, Azure OpenAI, Vertex AI or Ollama.

    For the complete list of models, please refer to the providers documentation.

    **Using OpenAI models**

    ```python
    import synalinks
    import os

    os.environ["OPENAI_API_KEY"] = "your-api-key"

    embedding_model = synalinks.EmbeddingModel(
        model="openai/text-embedding-ada-002",
    )
    ```

    **Using Azure OpenAI models**

    ```python
    import synalinks
    import os

    os.environ["AZURE_API_KEY"] = "your-api-key"
    os.environ["AZURE_API_BASE"] = "your-api-base"
    os.environ["AZURE_API_VERSION"] = "your-api-version"

    embedding_model = synalinks.EmbeddingModel(
        model="azure/<your_deployment_name>",
    )
    ```

    **Using VertexAI models**

    ```python
    import synalinks
    import os

    embedding_model = synalinks.EmbeddingModel(
        model="vertex_ai/text-embedding-004",
        vertex_project = "hardy-device-38811", # Your Project ID
        vertex_location = "us-central1",  # Project location
    )
    ```

    **Using Ollama models**

    ```python
    import synalinks

    embedding_model = synalinks.EmbeddingModel(
        model="ollama/mxbai-embed-large",
    )
    ```

    **Note**: Obviously, use an `.env` file and `.gitignore` to avoid
    putting your API keys in the code or a config file that can lead to
    leackage when pushing it into repositories.

    Args:
        model (str): The model to use.
        api_base (str): Optional. The endpoint to use.
        retry (int): Optional. The number of retry.
        fallback (EmbeddingModel): Optional. The embedding model to fallback
            if anything is wrong.
        caching (bool): Enables caching (Default to True).
    """

    def __init__(
        self,
        model=None,
        api_base=None,
        retry=5,
        fallback=None,
        caching=True,
    ):
        if model is None:
            raise ValueError(
                "You need to set the `model` argument for any EmbeddingModel"
            )
        self.model = model
        if self.model.startswith("ollama") and not api_base:
            self.api_base = "http://localhost:11434"
        else:
            self.api_base = api_base
        self.retry = retry
        self.fallback = fallback
        self.caching = caching

    async def __call__(self, texts, **kwargs):
        """
        Call method to get dense embeddings vectors

        Args:
            texts (list): A list of texts to embed.

        Returns:
            (list): The list of corresponding vectors.
        """

        for i in range(self.retry):
            try:
                if self.api_base:
                    response = await litellm.aembedding(
                        model=self.model,
                        input=texts,
                        api_base=self.api_base,
                        caching=self.caching,
                        **kwargs,
                    )
                else:
                    response = await litellm.aembedding(
                        model=self.model,
                        input=texts,
                        caching=self.caching,
                        **kwargs,
                    )
                vectors = []
                for data in response["data"]:
                    vectors.append(data["embedding"])
                return {"embeddings": vectors}
            except Exception as e:
                warnings.warn(f"Error occured while trying to call {self}: " + str(e))
        if self.fallback:
            return self.fallback(
                texts,
                **kwargs,
            )
        else:
            return None

    def _obj_type(self):
        return "EmbeddingModel"

    def get_config(self):
        config = {
            "model": self.model,
            "api_base": self.api_base,
            "retry": self.retry,
        }
        if self.fallback:
            fallback_config = {
                "fallback": serialization_lib.serialize_synalinks_object(
                    self.fallback,
                )
            }
            return {**fallback_config, **config}
        else:
            return config

    @classmethod
    def from_config(cls, config):
        if "fallback" in config:
            fallback = serialization_lib.deserialize_synalinks_object(
                config.pop("fallback")
            )
            return cls(fallback=fallback, **config)
        else:
            return cls(**config)

    def __repr__(self):
        api_base = f" api_base={self.api_base}" if self.api_base else ""
        return f"<EmbeddingModel model={self.model}{api_base}>"



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)
import copy


def dynamic_enum(schema, prop_to_update, labels, description=None):
    """Update a schema with dynamic Enum string.

    Args:
        schema (dict): The schema to update.
        prop_to_update (str): The property to update.
        labels (list): The list of labels (strings).
        description (str, optional): An optional description for the enum.

    Returns:
        dict: The updated schema with the enum applied to the specified property.
    """
    schema = copy.deepcopy(schema)
    if schema.get("$defs"):
        schema = {"$defs": schema.pop("$defs"), **schema}
    else:
        schema = {"$defs": {}, **schema}
    title = prop_to_update.title().replace("_", "")

    if description:
        enum_definition = {
            "enum": labels,
            "description": description,
            "title": title,
            "type": "string",
        }
    else:
        enum_definition = {
            "enum": labels,
            "title": title,
            "type": "string",
        }

    schema["$defs"].update({title: enum_definition})

    schema.setdefault("properties", {}).update(
        {prop_to_update: {"$ref": f"#/$defs/{title}"}}
    )

    return schema


def dynamic_tool_calls(tools):
    """
    Generates a dynamic schema for tool calls based on a list of tools.

    This function takes a list of tool objects and constructs a schema that includes
    definitions for each tool's properties, ensuring that each tool call includes a
    "tool_name" field to identify the tool being called.

    Args:
        tools (list): A list of tool objects, each with a name() method and
            an obj_schema() method that returns the schema of the tool.

    Returns:
        (dict): A schema dictionary that defines the structure for tool calls. The schema
            includes definitions for each tool and specifies that tool calls should
            be an array of items, each adhering to one of the tool schemas. The schema
            enforces that the "tool_name" field is required for each tool call.
    """
    tools_schemas_with_tool_names = {}

    for tool in tools:
        tool_name = tool.name
        schema = copy.deepcopy(tool.get_tool_schema())
        if "properties" in schema:
            tool_name_property = {
                "const": tool_name,
                "title": "Tool Name",
                "type": "string",
            }
            new_properties = {"tool_name": tool_name_property, **schema["properties"]}
            schema["properties"] = new_properties

        if "required" in schema:
            required_fields = ["tool_name"]
            required_fields.extend([req for req in schema["required"]])
            schema["required"] = required_fields
        else:
            schema["required"] = ["tool_name"]
        tools_schemas_with_tool_names[schema["title"]] = schema

    tool_calls_schema = {
        "$defs": tools_schemas_with_tool_names,
        "additionalProperties": False,
        "properties": {
            "tool_calls": {
                "items": {
                    "anyOf": [
                        {"$ref": "#/$defs/" + schema_key}
                        for schema_key in tools_schemas_with_tool_names.keys()
                    ]
                },
                "title": "Tool Calls",
                "type": "array",
            }
        },
        "required": ["tool_calls"],
        "title": "ToolCalls",
        "type": "object",
    }

    return tool_calls_schema


def dynamic_tool_choice(tools):
    tools_schemas_with_tool_names = {}

    for tool in tools:
        tool_name = tool.name
        schema = copy.deepcopy(tool.get_tool_schema())
        if "properties" in schema:
            tool_name_property = {
                "const": tool_name,
                "title": "Tool Name",
                "type": "string",
            }
            new_properties = {"tool_name": tool_name_property, **schema["properties"]}
            schema["properties"] = new_properties

        if "required" in schema:
            required_fields = ["tool_name"]
            required_fields.extend([req for req in schema["required"]])
            schema["required"] = required_fields
        else:
            schema["required"] = ["tool_name"]
        tools_schemas_with_tool_names[schema["title"]] = schema

    tool_choice_schema = {
        "$defs": tools_schemas_with_tool_names,
        "additionalProperties": False,
        "properties": {
            "tool_choice": {
                "anyOf": [
                    {"$ref": "#/$defs/" + schema_key}
                    for schema_key in tools_schemas_with_tool_names.keys()
                ],
                "title": "Tool Choice",
            }
        },
        "required": ["tool_choice"],
        "title": "ToolChoice",
        "type": "object",
    }

    return tool_choice_schema



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

import synalinks
from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules import Input
from synalinks.src.modules.core.action import Action
from synalinks.src.modules.core.tool import Tool
from synalinks.src.programs import Program


class ActionModuleTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_basic_action(self, mock_completion):
        class Query(DataModel):
            query: str

        @synalinks.saving.register_synalinks_serializable()
        async def calculate(expression: str):
            """Calculate the result of a mathematical expression.

            Args:
                expression (str): The mathematical expression to calculate, such as
                    '2 + 2'. The expression can contain numbers, operators
                    (+, -, *, /), parentheses and spaces.
            """
            if not all(char in "0123456789+-*/(). " for char in expression):
                return {
                    "result": None,
                    "log": "Error: invalid characters in expression",
                }
            try:
                # Evaluate the mathematical expression safely
                result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
                return {
                    "result": result,
                    "log": "Successfully executed",
                }
            except Exception as e:
                return {
                    "result": None,
                    "log": f"Error: {e}",
                }

        language_model = LanguageModel("ollama_chat/mistral")

        expected_string = """{"expression": "12 + 15"}"""

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        x0 = Input(data_model=Query)
        x1 = await Action(
            tool=Tool(calculate),
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="calculator",
            description=(
                "This program perform the calculation of a mathematical expression"
            ),
        )

        result = await program(
            Query(
                query=(
                    "You have a basket with 12 apples. "
                    "Your friend gives you 15 more apples. "
                    "How many apples do you have in total now?"
                )
            )
        )

        expected_json = {
            "action": {
                "inputs": {
                    "expression": "12 + 15",
                },
                "outputs": {
                    "result": 27.00,
                    "log": "Successfully executed",
                },
            }
        }

        self.assertEqual(result.get_json(), expected_json)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import collections
import copy

from synalinks.src.utils.nlp_utils import add_suffix
from synalinks.src.utils.nlp_utils import is_plural
from synalinks.src.utils.nlp_utils import to_plural_without_numerical_suffix
from synalinks.src.utils.nlp_utils import to_singular_without_numerical_suffix


def standardize_schema(schema):
    """Standardize the JSON schema for consistency"""
    return schema


def contains_schema(schema1, schema2):
    """Returns True if schema2 properties is a subset of schema1 properties."""
    return schema2.get("properties").items() <= schema1.get("properties").items()


def _default_title_from_key(key: str) -> str:
    return key.replace("_", " ").title()


def prefix_schema(schema, prefix):
    """Add a prefix to the schema properties"""
    schema = copy.deepcopy(schema)
    new_properties = {}

    for prop_key, prop_value in schema.get("properties", {}).items():
        prop_value = copy.deepcopy(prop_value)
        title = prop_value.get("title", _default_title_from_key(prop_key))
        prop_value["title"] = f"{prefix.title()} {title}"
        new_properties[f"{prefix}_{prop_key}"] = prop_value
    schema["properties"] = new_properties
    schema["required"] = list(new_properties.keys())
    return schema


def suffix_schema(schema, suffix):
    """Add a suffix to the schema properties"""
    schema = copy.deepcopy(schema)
    new_properties = {}

    for prop_key, prop_value in schema.get("properties", {}).items():
        prop_value = copy.deepcopy(prop_value)
        title = prop_value.get("title", _default_title_from_key(prop_key))
        prop_value["title"] = f"{title} {suffix.title()}"
        new_properties[f"{prop_key}_{suffix}"] = prop_value
    schema["properties"] = new_properties
    schema["required"] = list(new_properties.keys())
    return schema


def is_schema_equal(schema1, schema2):
    """Check if two JSON schemas are equal based on their properties.

    This function compares the properties of two JSON schemas
    to determine if they are equal. It only considers the properties
    and ignores other aspects of the schemas.

    Args:
        schema1 (dict): The first JSON schema to be compared.
        schema2 (dict): The second JSON schema to be compared.

    Returns:
        - (bool): `True` if the schemas are considered equal based on their properties,
        `False` otherwise.
    """
    schema1_properties = schema1.get("properties", {})
    schema2_properties = schema2.get("properties", {})
    if len(schema1_properties) != len(schema2_properties):
        return False
    return schema1_properties == schema2_properties


def is_object(schema):
    """Returns true if the schema is an object's schema"""
    if isinstance(schema, dict):
        if schema.get("type", None) == "object":
            return True
    return False


def is_array(schema):
    """Returns true if the schema is an array's schema"""
    if isinstance(schema, dict):
        if schema.get("type", None) == "array":
            return True
    return False


def concatenate_schema(schema1, schema2):
    """Concatenate two JSON schemas into a single schema.

    This function merges the properties of two JSON schemas into a single schema.
    If there are conflicting property names, it appends a suffix to make them unique.

    Args:
        schema1 (dict): The first JSON schema to be concatenated.
        schema2 (dict): The second JSON schema to be concatenated.

    Returns:
        (dict): A new JSON schema that combines the properties of the input schemas.
    """
    schema1 = copy.deepcopy(schema1)
    schema2 = copy.deepcopy(schema2)
    # Initialize the resulting schema
    result_schema = {
        "additionalProperties": False,
        "$defs": {},
        "properties": {},
        "required": [],
        "title": schema1.get("title"),
        "type": "object",
    }

    if schema1.get("$defs") and not schema2.get("$defs"):
        result_schema["$defs"] = schema1.get("$defs")
    if not schema1.get("$defs") and schema2.get("$defs"):
        result_schema["$defs"] = schema2.get("$defs")
    if schema1.get("$defs") and schema2.get("$defs"):
        result_schema["$defs"] = {**schema1.get("$defs"), **schema2.get("$defs")}

    schema1_properties = schema1.get("properties", {})
    schema2_properties = schema2.get("properties", {})

    # Helper function to add a property to the result schema
    def add_property(prop_key, prop_value, suffix=0):
        new_prop_key = prop_key
        while new_prop_key in result_schema["properties"]:
            suffix += 1
            new_prop_key = add_suffix(prop_key, suffix)
            prop_value["title"] = new_prop_key.title().replace("_", " ")
        result_schema["properties"][new_prop_key] = prop_value

        required1 = schema1.get("required")
        required2 = schema2.get("required")
        if required1 and not required2:
            if prop_key in required1:
                result_schema["required"].append(new_prop_key)

        if required2 and not required1:
            if prop_key in required2:
                result_schema["required"].append(new_prop_key)

        if required1 and required2:
            if prop_key in required1 or prop_key in required2:
                result_schema["required"].append(new_prop_key)

    # Add properties from the first schema
    for prop_key, prop_value in schema1_properties.items():
        add_property(prop_key, prop_value)

    # Add properties from the second schema
    for prop_key, prop_value in schema2_properties.items():
        add_property(prop_key, prop_value)

    if len(result_schema.get("$defs")) == 0:
        del result_schema["$defs"]

    return result_schema


def factorize_schema(schema):
    """Factorize a JSON schema by grouping similar properties into lists.

    This function groups similar properties in a JSON schema into list properties.
        It identifies similar properties based on their base names
        and creates array properties for them. The grouped properties are renamed in
        their plural form.

    Args:
        schema (dict): The input JSON schema to factorize.

    Returns:
        (dict): A factorized JSON schema with grouped properties.
    """
    schema = copy.deepcopy(schema)
    # Initialize the resulting schema
    result_schema = {
        "$defs": {},
        "additionalProperties": False,
        "properties": {},
        "required": [],
        "title": schema.get("title"),
        "type": "object",
    }

    if schema.get("$defs"):
        result_schema["$defs"] = schema.get("$defs")

    schema_properties = schema.get("properties", {})

    for prop_key, prop_value in schema_properties.items():
        # Get the base name
        base_key = to_singular_without_numerical_suffix(prop_key)
        plural_key = to_plural_without_numerical_suffix(base_key)
        # Find all similar properties
        similar_prop_keys = [
            p
            for p in schema_properties.keys()
            if to_singular_without_numerical_suffix(p) == base_key and p != prop_key
        ]
        similar_prop_values = [schema["properties"][p] for p in similar_prop_keys]
        if similar_prop_keys and not is_plural(prop_key):
            if plural_key not in result_schema["properties"]:
                # Create an array property
                array_prop = copy.deepcopy(prop_value)
                array_prop["title"] = plural_key.title()
                array_prop["type"] = "array"
                if is_array(prop_value):
                    if all(is_array(prop) for prop in similar_prop_values):
                        if not all(
                            prop["items"] == prop_value["items"]
                            for prop in similar_prop_values
                        ):
                            if "$ref" in array_prop["items"]:
                                del array_prop["items"]["$ref"]

                            for s in similar_prop_values:
                                if s["items"] != prop_value["items"]:
                                    if "anyOf" not in array_prop["items"]:
                                        array_prop["items"]["anyOf"] = []
                                    array_prop["items"]["anyOf"].append(s["items"])
                            array_prop["items"]["anyOf"].append(prop_value["items"])
                            if "description" in array_prop:
                                del array_prop["description"]
                        else:
                            array_prop["items"] = prop_value["items"]
                    else:
                        array_prop["items"] = prop_value["items"]
                else:
                    array_prop["items"] = {"type": prop_value["type"]}

                result_schema["properties"][plural_key] = array_prop
                if plural_key not in result_schema["required"]:
                    result_schema["required"].append(plural_key)
        else:
            if not is_plural(prop_key):
                result_schema["properties"][base_key] = prop_value
                if base_key not in result_schema["required"]:
                    result_schema["required"].append(base_key)

    if len(result_schema.get("$defs")) == 0:
        del result_schema["$defs"]

    return result_schema


def out_mask_schema(schema, mask=None, recursive=True):
    """Mask specific fields of a JSON schema.

    This function looks for properties to mask and removes them.
    It ignores the suffixes that other operations could add.

    Args:
        schema (dict): The input JSON schema to mask.
        mask (list): The base key list to remove.
        recursive (bool): Whether or not to remove
            recursively for nested objects (default True).

    Returns:
        (dict): A masked JSON schema with removed properties.
    """
    schema = copy.deepcopy(schema)

    if not mask:
        return schema

    stack_init = [schema]

    if recursive:
        if "$defs" in schema:
            for obk_name, obj_schema in schema["$defs"].items():
                stack_init.append(obj_schema)

    stack = collections.deque(stack_init)

    # Ensure that the mask keys are in singular form
    mask = [to_singular_without_numerical_suffix(k) for k in mask]

    while stack:
        current = stack.pop()
        keys_to_delete = []

        properties = current.get("properties", {})

        for prop_key, prop_value in properties.items():
            base_key = to_singular_without_numerical_suffix(prop_key)
            if base_key in mask:
                keys_to_delete.append(prop_key)

            if recursive:
                if is_object(prop_value):
                    stack.append(prop_value)
                elif is_array(prop_value):
                    stack.append(prop_value["items"])

        for key in keys_to_delete:
            del properties[key]

        if "required" in current:
            temp_required = []
            for req in current.get("required"):
                if req not in keys_to_delete:
                    temp_required.append(req)
            current["required"] = temp_required

    # Clean up defs if no link found after masking
    new_defs = {}
    if "$defs" in schema:
        for obj_key, obj_schema in schema["$defs"].items():
            if str(schema).find(f"#/$defs/{obj_key}") > 0:
                new_defs[obj_key] = obj_schema
        schema["$defs"] = new_defs

    return schema


def in_mask_schema(schema, mask=None, recursive=True):
    """Keep specific fields of a JSON schema.

    This function looks for properties to keep and removes all others.
    It ignores the suffixes that other operations could add.

    Args:
        - schema (dict): The input JSON schema to mask.
        - mask (list): The base key list to keep.
        - recursive (bool): Whether or not to keep
            recursively for nested objects (default True).

    Returns:
        - (dict): A masked JSON schema with only the specified properties.
    """
    schema = copy.deepcopy(schema)

    if not mask:
        return {
            "additionalProperties": False,
            "properties": {},
            "title": schema.get("title"),
            "type": "object",
        }

    stack_init = [schema]

    if recursive:
        if "$defs" in schema:
            for _, obj_schema in schema["$defs"].items():
                stack_init.append(obj_schema)

    stack = collections.deque(stack_init)

    # Ensure that the mask keys are in singular form
    mask = [to_singular_without_numerical_suffix(k) for k in mask]

    while stack:
        current = stack.pop()
        keys_to_keep = []

        properties = current.get("properties", {})

        for prop_key, prop_value in properties.items():
            base_key = to_singular_without_numerical_suffix(prop_key)

            if base_key in mask:
                keys_to_keep.append(prop_key)

            if recursive:
                if is_object(prop_value):
                    stack.append(prop_value)
                elif is_array(prop_value):
                    stack.append(prop_value["items"])

        keys_to_delete = set(properties.keys()) - set(keys_to_keep)
        for key in keys_to_delete:
            del properties[key]
        if "required" in current:
            required = []
            for req in current["required"]:
                if req in keys_to_keep:
                    required.append(req)
            if required:
                current["required"] = required
            else:
                del current["required"]

    # Clean up defs if no link found after masking
    new_defs = {}
    if "$defs" in schema:
        for obj_key, obj_schema in schema["$defs"].items():
            if str(schema).find(f"#/$defs/{obj_key}") > 0:
                new_defs[obj_key] = obj_schema
        schema["$defs"] = new_defs

    return schema



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import logging

import orjson

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.hooks.hook import Hook

_SYMBOLIC_LOG_TEMPLATE = """
Call ID: {call_id}
Parent call ID: {parent_call_id}
Module: {module}
Module Name: {module_name}
Module Description: {module_description}
Data Model JSON Schema:
{data_model_schema}
"""

_DATA_LOG_TEMPLATE = """
Call ID: {call_id}
Parent call ID: {parent_call_id}
Module: {module}
Module Name: {module_name}
Module Description: {module_description}
Data Model JSON:
{data_model_json}
"""

_EXCEPTION_TEMPLATE = """
# Exception
Call ID: {call_id}
Parent call ID: {parent_call_id}
Module: {module}
Module Name: {module_name}
Module Description: {module_description}
Exception: {exception}
"""

_KWARGS_LOG_TEMPLATE = """
Call ID: {call_id}
Parent call ID: {parent_call_id}
Module: {module}
Module Name: {module_name}
Module Description: {module_description}
Keyword Arguments:
{kwargs_json}
"""


@synalinks_export("synalinks.hooks.Logger")
class Logger(Hook):
    """Logger hook for logging module calls.

    This hook is set by default when you enables logging.

    Example:

    ```python
    import synalinks

    synalinks.enable_logging()
    ```
    """

    def _maybe_setup_logger(self):
        if not hasattr(self, "logger"):
            self.logger = logging.getLogger(f"synalinks.{self.module.name}")

    def _serialize_kwargs(self, kwargs):
        """Serialize kwargs to JSON-compatible format."""
        if not kwargs:
            return {}
        serialized = {}
        for key, value in kwargs.items():
            if key == "training":
                serialized[key] = value
            elif hasattr(value, "get_json"):
                serialized[key] = value.get_json()
            elif isinstance(value, (str, int, float, bool, list, dict, type(None))):
                serialized[key] = value
        return serialized

    def on_call_begin(
        self,
        call_id,
        parent_call_id=None,
        inputs=None,
        kwargs=None,
    ):
        self._maybe_setup_logger()
        module_name = self.module.name
        module_description = self.module.description
        flatten_inputs = tree.flatten(inputs)
        if any_symbolic_data_models(inputs):
            data_models_schemas = [
                dm.get_schema() for dm in flatten_inputs if dm is not None
            ]
            if data_models_schemas:
                self.logger.debug(
                    _SYMBOLIC_LOG_TEMPLATE.format(
                        name="Symbolic Call Start",
                        call_id=call_id,
                        parent_call_id=parent_call_id,
                        module=str(self.module.__class__.__name__),
                        module_name=module_name,
                        module_description=module_description,
                        data_model_schema=orjson.dumps(
                            data_models_schemas,
                            option=orjson.OPT_INDENT_2,
                        ).decode(),
                    ),
                )
        else:
            data_models_jsons = [dm.get_json() for dm in flatten_inputs if dm is not None]
            if data_models_jsons:
                self.logger.info(
                    _DATA_LOG_TEMPLATE.format(
                        name="Call Start",
                        call_id=call_id,
                        parent_call_id=parent_call_id,
                        module=str(self.module.__class__.__name__),
                        module_name=module_name,
                        module_description=module_description,
                        data_model_json=orjson.dumps(
                            data_models_jsons,
                            option=orjson.OPT_INDENT_2,
                        ).decode(),
                    )
                )
            # Log kwargs if no data models but kwargs present (e.g., Tool modules)
            elif kwargs:
                serialized_kwargs = self._serialize_kwargs(kwargs)
                # Filter out 'training' from display
                display_kwargs = {
                    k: v for k, v in serialized_kwargs.items() if k != "training"
                }
                if display_kwargs:
                    self.logger.info(
                        _KWARGS_LOG_TEMPLATE.format(
                            call_id=call_id,
                            parent_call_id=parent_call_id,
                            module=str(self.module.__class__.__name__),
                            module_name=module_name,
                            module_description=module_description,
                            kwargs_json=orjson.dumps(
                                display_kwargs,
                                option=orjson.OPT_INDENT_2,
                            ).decode(),
                        )
                    )

    def on_call_end(
        self,
        call_id,
        parent_call_id=None,
        outputs=None,
        exception=None,
    ):
        self._maybe_setup_logger()
        module_name = self.module.name
        module_description = self.module.description
        if exception:
            self.logger.error(
                _EXCEPTION_TEMPLATE.format(
                    call_id=call_id,
                    parent_call_id=parent_call_id,
                    exception=exception,
                    module=str(self.module.__class__.__name__),
                    module_name=module_name,
                    module_description=module_description,
                )
            )
        if not outputs:
            return
        flatten_outputs = tree.flatten(outputs)
        if any_symbolic_data_models(outputs):
            data_models_schemas = [
                dm.get_schema() for dm in flatten_outputs if dm is not None
            ]
            if data_models_schemas:
                self.logger.debug(
                    _SYMBOLIC_LOG_TEMPLATE.format(
                        name="Symbolic Call End",
                        call_id=call_id,
                        parent_call_id=parent_call_id,
                        module=str(self.module.__class__.__name__),
                        module_name=module_name,
                        module_description=module_description,
                        data_model_schema=orjson.dumps(
                            data_models_schemas,
                            option=orjson.OPT_INDENT_2,
                        ).decode(),
                    ),
                )
        else:
            data_models_jsons = [
                dm.get_json() for dm in flatten_outputs if dm is not None
            ]
            if data_models_jsons:
                self.logger.info(
                    _DATA_LOG_TEMPLATE.format(
                        name="Call End",
                        call_id=call_id,
                        parent_call_id=parent_call_id,
                        module=str(self.module.__class__.__name__),
                        module_name=module_name,
                        module_description=module_description,
                        data_model_json=orjson.dumps(
                            data_models_jsons,
                            option=orjson.OPT_INDENT_2,
                        ).decode(),
                    )
                )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.rewards import deserialize as deserialize
from synalinks.src.rewards import get as get
from synalinks.src.rewards import serialize as serialize
from synalinks.src.rewards.cosine_similarity import CosineSimilarity as CosineSimilarity
from synalinks.src.rewards.cosine_similarity import cosine_similarity as cosine_similarity
from synalinks.src.rewards.exact_match import ExactMatch as ExactMatch
from synalinks.src.rewards.exact_match import exact_match as exact_match
from synalinks.src.rewards.lm_as_judge import LMAsJudge as LMAsJudge
from synalinks.src.rewards.reward import Reward as Reward
from synalinks.src.rewards.reward_wrappers import ProgramAsJudge as ProgramAsJudge
from synalinks.src.rewards.reward_wrappers import (
    RewardFunctionWrapper as RewardFunctionWrapper,
)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import copy
import sys
import traceback
from io import StringIO
from multiprocessing import Process
from multiprocessing import Queue

import jsonschema
from jsonschema import ValidationError

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import Trainable
from synalinks.src.modules.module import Module


class TimeoutException(Exception):
    """Exception raised when script execution times out"""

    pass


class PythonScript(Trainable):
    """The python code to transform a JSON object into another JSON object"""

    python_script: str = Field(
        description="The python script to transform a JSON object into another object"
    )


class PythonConsoleLog(DataModel):
    stdout: str = Field(description="The python console's stdout")
    stderr: str = Field(description="The python console's stderr")


def _execute_script_in_process(python_script, inputs_json, schema, result_queue):
    """Execute the script in a separate process.

    This function runs in a separate process and can be forcefully terminated.
    """
    result = None

    # Capture stdout and stderr
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    stdout_capture = StringIO()
    stderr_capture = StringIO()
    sys.stdout = stdout_capture
    sys.stderr = stderr_capture

    try:
        # Create a local namespace with the inputs
        local_namespace = {"inputs": copy.deepcopy(inputs_json)}

        # Execute the entire script
        exec(python_script, local_namespace)

        # Look for the result variable in the namespace
        if "result" in local_namespace:
            result = local_namespace["result"]

            if result:
                try:
                    jsonschema.validate(result, schema)
                except ValidationError as validation_error:
                    stdout = stdout_capture.getvalue()
                    stderr = (
                        stderr_capture.getvalue()
                        + f"Validation Error: {validation_error}\n"
                    )
                    result_queue.put((None, stdout, stderr))
                    return
        else:
            stdout = stdout_capture.getvalue()
            stderr = (
                stderr_capture.getvalue()
                + "Error: No 'result' variable found after script execution\n"
            )
            result_queue.put((None, stdout, stderr))
            return

    except Exception as e:
        stdout = stdout_capture.getvalue()
        stderr = stderr_capture.getvalue() + f"Error: {str(e)}\n{traceback.format_exc()}"
        result_queue.put((None, stdout, stderr))
        return
    finally:
        stdout = stdout_capture.getvalue()
        stderr = stderr_capture.getvalue()
        sys.stdout = old_stdout
        sys.stderr = old_stderr

    result_queue.put((result, stdout, stderr))


@synalinks_export(
    [
        "synalinks.modules.PythonSynthesis",
        "synalinks.PythonSynthesis",
    ]
)
class PythonSynthesis(Module):
    """A code Python code transformation on JSON data.

    **Note**: This module is **NOT** completly safe (yet) for business applications.
        Its is only provided for reseach purposes on program synthesis.
        Altought the code don't evolve during inference, so it can't be prompt injected.

    This module features a python code as trainable variable, allowing the optimizers
    to refine the code during the training loop based on iterative feedback and
    automatic selection of the best script.

    This module works **ONLY** with advanced optimizers (**NOT** the
    `RandomFewShot` optimizer).

    The module executes the entire Python script and expects the result to be stored
    in a variable named 'result' at the end of execution.
    
    Example:
    
    ```python
    import synalinks
    import asyncio
    
    default_python_script = \\
    \"\"\"
    def transform(inputs):
        # TODO implement the code to transform the input grid into the output grid
        return {"output_grid": inputs.get("input_grid")}
        
    result = transform(inputs)
    \"\"\"
    
    async def main():
        inputs = synalinks.Input(
            data_model=synalinks.datasets.arcagi.get_input_data_model(),
        )
        outputs = await synalinks.PythonSynthesis(
            data_model=synalinks.datasets.arcagi.get_output_data_model()
            python_script=default_python_script,
            default_return_value={"output_grid": [[]]},
        )(inputs)
        
        program = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="python_script_synthesis",
            description="A program to solve ARCAGI with python code",
        )
    ```
    
    If you want to explore the future of neuro-symbolic self-evolving systems, contact us.
    While these systems are not "hard" to code thanks to Synalinks, they requires 
    technical knowledge and a deep understanding of multiple AI paradigm.

    Args:
        schema (dict): The target JSON schema.
            If not provided use the `data_model` to infer it.
        data_model (DataModel | SymbolicDataModel | JsonDataModel): The target data
            model for structured output.
        python_script (str): The default Python script.
        seed_scripts (list): Optional. A list of Python scripts to use as seed
            for the evolution. If not provided, create a seed from the default
            configuration.
        default_return_value (dict): Default return value.
        return_python_script (bool): Wether or not to return the python script for 
            evaluation. (Default to False).
        timeout (int): Maximum execution time in seconds. (Default 5 seconds).
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        schema=None,
        data_model=None,
        python_script=None,
        seed_scripts=None,
        default_return_value=None,
        return_python_script=False,
        timeout=5,
        sandbox=False,
        name=None,
        description=None,
        trainable=True,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        if not schema and data_model:
            schema = data_model.get_schema()
        self.schema = schema

        if not python_script:
            raise ValueError("You should provide the `python_script` argument")
        self.python_script = python_script

        if not default_return_value:
            raise ValueError("You should provide the `default_return_value` argument")

        try:
            jsonschema.validate(default_return_value, self.schema)
        except ValidationError as e:
            raise ValueError(
                f"`default_return_value` parameter does not conform to schema: {e}"
            )

        self.default_return_value = default_return_value
        self.return_python_script = return_python_script
        self.timeout = timeout

        if not seed_scripts:
            seed_scripts = []
        self.seed_scripts = seed_scripts

        seed_candidates = [
            {"python_script": seed_script} for seed_script in self.seed_scripts
        ]

        self.state = self.add_variable(
            initializer=PythonScript(
                python_script=self.python_script,
                seed_candidates=seed_candidates,
            ).get_json(),
            data_model=PythonScript,
            name="state_" + self.name,
        )

    async def execute(self, inputs, python_script):
        """Execute the Python script with timeout using multiprocessing."""
        result_queue = Queue()

        process = Process(
            target=_execute_script_in_process,
            args=(python_script, inputs.get_json(), self.schema, result_queue),
        )
        process.start()

        start_time = asyncio.get_event_loop().time()
        timeout_remaining = self.timeout

        while process.is_alive() and timeout_remaining > 0:
            await asyncio.sleep(0.1)
            elapsed = asyncio.get_event_loop().time() - start_time
            timeout_remaining = self.timeout - elapsed

        if process.is_alive():
            process.terminate()
            process.join(timeout=1)

            if process.is_alive():
                process.kill()
                process.join()

            return (
                None,
                "",
                f"Timeout Error: Script execution exceeded {self.timeout} second(s)\n",
            )

        process.join()

        if not result_queue.empty():
            result, stdout, stderr = result_queue.get()
            return result, stdout, stderr
        else:
            return None, "", "Error: Process terminated unexpectedly\n"

    async def call(self, inputs, training=False):
        if not inputs:
            return None
        python_script = self.state.get("python_script")
        result, stdout, stderr = await self.execute(inputs, python_script)
        if training:
            predictions = self.state.get("current_predictions")
            if result:
                if self.return_python_script:
                    predictions.append(
                        {
                            "inputs": {
                                **inputs.get_json(),
                            },
                            "outputs": {
                                "python_script": python_script,
                                **result,
                                "stdout": stdout,
                                "stderr": stderr,
                            },
                            "reward": None,
                        }
                    )
                else:
                    predictions.append(
                        {
                            "inputs": {
                                **inputs.get_json(),
                            },
                            "outputs": {
                                **result,
                                "stdout": stdout,
                                "stderr": stderr,
                            },
                            "reward": None,
                        }
                    )
            else:
                if self.return_python_script:
                    predictions.append(
                        {
                            "inputs": {
                                **inputs.get_json(),
                            },
                            "outputs": {
                                "python_script": python_script,
                                "stdout": stdout,
                                "stderr": stderr,
                            },
                            "reward": None,
                        }
                    )
                else:
                    predictions.append(
                        {
                            "inputs": {
                                **inputs.get_json(),
                            },
                            "outputs": {
                                "stdout": stdout,
                                "stderr": stderr,
                            },
                            "reward": None,
                        }
                    )
        if result:
            if self.return_python_script:
                return JsonDataModel(
                    json={
                        "python_script": python_script,
                        **result,
                        "stdout": stdout,
                        "stderr": stderr,
                    },
                    schema=self.schema,
                    name=self.name,
                )
            else:
                return JsonDataModel(
                    json={
                        **result,
                        "stdout": stdout,
                        "stderr": stderr,
                    },
                    schema=self.schema,
                    name=self.name,
                )
        else:
            if self.return_python_script:
                return JsonDataModel(
                    json={
                        "python_script": python_script,
                        **self.default_return_value,
                        "stdout": stdout,
                        "stderr": stderr,
                    },
                    schema=self.schema,
                    name=self.name,
                )
            else:
                return JsonDataModel(
                    json={
                        **self.default_return_value,
                        "stdout": stdout,
                        "stderr": stderr,
                    },
                    schema=self.schema,
                    name=self.name,
                )

    async def compute_output_spec(self, inputs, training=False):
        if self.return_python_script:
            return await ops.concat(
                await ops.out_mask(
                    PythonScript.to_symbolic_data_model(),
                    mask=list(Trainable.keys()),
                    name="python_script_masked_" + self.name,
                ),
                await ops.concat(
                    SymbolicDataModel(schema=self.schema),
                    PythonConsoleLog,
                    name="python_logs_" + self.name,
                ),
                name=self.name,
            )
        else:
            return await ops.concat(
                SymbolicDataModel(schema=self.schema),
                PythonConsoleLog,
                name=self.name,
            )

    def get_config(self):
        config = {
            "schema": self.schema,
            "python_script": self.python_script,
            "seed_scripts": self.seed_scripts,
            "default_return_value": self.default_return_value,
            "return_python_script": self.return_python_script,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Modified from: keras/src/backend/common/variables_test.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import List

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Variable
from synalinks.src.backend import standardize_schema


class VariablesTest(testing.TestCase):
    def test_initialize_variable_with_dict(self):
        class Instructions(DataModel):
            instructions: List[str] = []

        initial_data = {
            "instructions": [
                "For any problem involving division, always round the quotient to "
                "the nearest even number, regardless of the remainder."
            ],
        }
        variable_from_dict = Variable(
            initializer=initial_data,
            data_model=Instructions,
        )
        self.assertEqual(variable_from_dict.get_json(), initial_data)
        self.assertEqual(
            variable_from_dict.get_schema(),
            standardize_schema(Instructions.get_schema()),
        )

    def test_initialize_variable_with_callable_initializer(self):
        class Instructions(DataModel):
            instructions: List[str] = []

        from synalinks.src.initializers import Empty

        variable_from_initializer = Variable(initializer=Empty(data_model=Instructions))
        self.assertEqual(variable_from_initializer.get_json(), Instructions().get_json())
        self.assertEqual(
            variable_from_initializer.get_schema(),
            standardize_schema(Instructions.get_schema()),
        )

    def test_assign_variable_from_dict(self):
        class Instructions(DataModel):
            instructions: List[str] = []

        initial_data = {
            "instructions": [
                "For any problem involving division, always round the quotient to "
                "the nearest even number, regardless of the remainder."
            ],
        }
        variable_from_dict = Variable(initializer=initial_data, data_model=Instructions)
        new_value = {
            "instructions": [
                "When performing division, always check if the division results in "
                "a whole number. If not, express the result as a fraction or a "
                "decimal, depending on the context of the problem."
            ],
        }
        variable_from_dict.assign(new_value)
        self.assertEqual(variable_from_dict.get_json(), new_value)
        self.assertEqual(
            variable_from_dict.get_schema(),
            standardize_schema(Instructions.get_schema()),
        )

    def test_assign_variable_from_dataype(self):
        class Instructions(DataModel):
            instructions: List[str] = []

        initial_data = {
            "instructions": [
                "For any problem involving division, always round the quotient to "
                "the nearest even number, regardless of the remainder."
            ],
        }
        variable_from_dict = Variable(initializer=initial_data, data_model=Instructions)
        new_value = {
            "instructions": [
                "When performing division, always check if the division results in "
                "a whole number. If not, express the result as a fraction or a "
                "decimal, depending on the context of the problem."
            ],
        }
        variable_from_dict.assign(new_value)
        self.assertEqual(variable_from_dict.get_json(), new_value)
        self.assertEqual(
            variable_from_dict.get_schema(),
            standardize_schema(Instructions.get_schema()),
        )

    def test_contains_operator(self):
        class Foo(DataModel):
            foo: str = ""

        class FooBar(DataModel):
            foo: str = ""
            bar: str = ""

        variable = Variable(
            initializer={"foo": "value", "bar": "value2"}, data_model=FooBar
        )

        self.assertTrue(Foo in variable)
        self.assertFalse(FooBar in Foo)

    def test_contains_string_key_operator(self):
        class FooBar(DataModel):
            foo: str = ""
            bar: str = ""

        variable = Variable(
            initializer={"foo": "value", "bar": "value2"}, data_model=FooBar
        )

        self.assertTrue("foo" in variable)
        self.assertTrue("bar" in variable)
        self.assertFalse("baz" in variable)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json

from synalinks.src import backend
from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.metrics.f_score_metrics import BinaryF1Score
from synalinks.src.metrics.f_score_metrics import BinaryFBetaScore
from synalinks.src.metrics.f_score_metrics import F1Score
from synalinks.src.metrics.f_score_metrics import FBetaScore
from synalinks.src.metrics.f_score_metrics import ListF1Score
from synalinks.src.metrics.f_score_metrics import ListFBetaScore


class FBetaScoreTest(testing.TestCase):
    async def test_same_field(self):
        class Answer(DataModel):
            answer: str

        y_pred = Answer(answer="Toulouse is the French city of aeronautics and space.")
        y_true = Answer(answer="Toulouse is the French city of aeronautics and space.")

        metric = FBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_field(self):
        class Answer(DataModel):
            answer: str

        y_pred = Answer(answer="Toulouse is the French city of aeronautics and space.")
        y_true = Answer(answer="Paris is the capital of France.")

        metric = FBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())


class F1ScoreTest(testing.TestCase):
    async def test_same_field(self):
        class Answer(DataModel):
            answer: str

        y_pred = Answer(answer="Toulouse is the French city of aeronautics and space.")
        y_true = Answer(answer="Toulouse is the French city of aeronautics and space.")

        metric = F1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_field(self):
        class Answer(DataModel):
            answer: str

        y_pred = Answer(answer="Toulouse is the French city of aeronautics and space.")
        y_true = Answer(answer="Paris is the capital of France.")

        metric = F1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())


class BinaryFBetaScoreTest(testing.TestCase):
    async def test_same_boolean_fields(self):
        class MultiLabels(DataModel):
            label: bool
            label_1: bool
            label_2: bool

        y_pred = MultiLabels(label=False, label_1=True, label_2=True)
        y_true = MultiLabels(label=False, label_1=True, label_2=True)

        metric = BinaryFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_boolean_fields(self):
        class MultiLabels(DataModel):
            label: bool
            label_1: bool
            label_2: bool

        y_pred = MultiLabels(label=True, label_1=True, label_2=False)
        y_true = MultiLabels(label=False, label_1=False, label_2=True)

        metric = BinaryFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())


class BinaryF1ScoreTest(testing.TestCase):
    async def test_same_boolean_fields(self):
        class MultiLabels(DataModel):
            label: bool
            label_1: bool
            label_2: bool

        y_pred = MultiLabels(label=False, label_1=True, label_2=True)
        y_true = MultiLabels(label=False, label_1=True, label_2=True)

        metric = BinaryF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_boolean_fields(self):
        class MultiLabels(DataModel):
            label: bool
            label_1: bool
            label_2: bool

        y_pred = MultiLabels(label=True, label_1=True, label_2=False)
        y_true = MultiLabels(label=False, label_1=False, label_2=True)

        metric = BinaryF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())

    async def test_reset_state(self):
        class MultiLabels(DataModel):
            label: bool
            label_1: bool
            label_2: bool

        y_pred = MultiLabels(label=False, label_1=True, label_2=True)
        y_true = MultiLabels(label=False, label_1=True, label_2=True)

        metric = BinaryF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())
        metric.reset_state()
        score = metric.result()
        self.assertEqual(score, 0.0)

    async def test_variable_serialization(self):
        class MultiLabels(DataModel):
            label: bool
            label_1: bool
            label_2: bool

        y_pred = MultiLabels(label=False, label_1=True, label_2=True)
        y_true = MultiLabels(label=False, label_1=True, label_2=True)
        metric = BinaryF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())
        state = metric.variables[0]
        # Try to dump it so we can test if the state is serializable
        _ = json.dumps(state.get_json())


class ListFBetaScoreTest(testing.TestCase):
    async def test_same_single_label(self):
        from typing import Literal

        class SingleLabel(DataModel):
            label: Literal["label_1", "label_2", "label_3"]

        y_pred = SingleLabel(label="label_1")
        y_true = SingleLabel(label="label_1")

        metric = ListFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        print(score)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_single_label(self):
        from typing import Literal

        class SingleLabel(DataModel):
            label: Literal["label_1", "label_2", "label_3"]

        y_pred = SingleLabel(label="label_1")
        y_true = SingleLabel(label="label_2")

        metric = ListFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())

    async def test_same_multi_label(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1", "label_2"])
        y_true = MultiLabel(labels=["label_1", "label_2"])

        metric = ListFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_partial_overlap_multi_label(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1", "label_2"])
        y_true = MultiLabel(labels=["label_2", "label_3"])

        metric = ListFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        # Should have partial overlap (label_2 matches)
        self.assertGreater(score, 0.0)
        self.assertLess(score, 1.0)

    async def test_no_overlap_multi_label(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1"])
        y_true = MultiLabel(labels=["label_2", "label_3"])

        metric = ListFBetaScore(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())

    async def test_retrieval_sources(self):
        from typing import List

        class AnswerWithSources(DataModel):
            sources: List[str]
            answer: str

        y_pred = AnswerWithSources(
            sources=["source1", "source2", "source3"], answer="This is an answer"
        )
        y_true = AnswerWithSources(
            sources=["source1", "source2"], answer="This is an answer"
        )

        # Test with in_mask to only evaluate sources
        metric = ListFBetaScore(average="weighted", in_mask=["sources"])
        score = await metric(y_true, y_pred)
        # Should have partial match (2 out of 3 predicted, 2 out of 2 true)
        self.assertGreater(score, 0.0)
        self.assertLess(score, 1.0)


class ListF1ScoreTest(testing.TestCase):
    async def test_same_single_label(self):
        from typing import Literal

        class SingleLabel(DataModel):
            label: Literal["label_1", "label_2", "label_3"]

        y_pred = SingleLabel(label="label_1")
        y_true = SingleLabel(label="label_1")

        metric = ListF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_single_label(self):
        from typing import Literal

        class SingleLabel(DataModel):
            label: Literal["label_1", "label_2", "label_3"]

        y_pred = SingleLabel(label="label_1")
        y_true = SingleLabel(label="label_2")

        metric = ListF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())

    async def test_same_multi_label(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1", "label_2", "label_3"])
        y_true = MultiLabel(labels=["label_1", "label_2", "label_3"])

        metric = ListF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_different_multi_label(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1"])
        y_true = MultiLabel(labels=["label_2", "label_3"])

        metric = ListF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 0.0, delta=3 * backend.epsilon())

    async def test_reset_state(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1", "label_2"])
        y_true = MultiLabel(labels=["label_1", "label_2"])

        metric = ListF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())
        metric.reset_state()
        score = metric.result()
        self.assertEqual(score, 0.0)

    async def test_variable_serialization(self):
        from typing import List
        from typing import Literal

        class MultiLabel(DataModel):
            labels: List[Literal["label_1", "label_2", "label_3"]]

        y_pred = MultiLabel(labels=["label_1", "label_2"])
        y_true = MultiLabel(labels=["label_1", "label_2"])

        metric = ListF1Score(average="weighted")
        score = await metric(y_true, y_pred)
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())
        state = metric.variables[0]
        # Try to dump it so we can test if the state is serializable
        _ = json.dumps(state.get_json())

    async def test_with_out_mask(self):
        from typing import List

        class AnswerWithSources(DataModel):
            sources: List[str]
            answer: str

        y_pred = AnswerWithSources(
            sources=["source1", "source2"], answer="Different answer"
        )
        y_true = AnswerWithSources(
            sources=["source1", "source2"], answer="This is an answer"
        )

        # Test with out_mask to exclude answer field
        metric = ListF1Score(average="weighted", out_mask=["answer"])
        score = await metric(y_true, y_pred)
        # Sources match perfectly, answer is excluded
        self.assertAlmostEqual(score, 1.0, delta=3 * backend.epsilon())

    async def test_averaging_modes(self):
        from typing import Literal

        class MultiField(DataModel):
            field1: Literal["a", "b", "c"]
            field2: Literal["x", "y", "z"]

        y_pred = MultiField(field1="a", field2="x")
        y_true = MultiField(field1="a", field2="y")

        # Test with no averaging
        metric_none = ListF1Score(average=None)
        score_none = await metric_none(y_true, y_pred)
        self.assertIsInstance(score_none, list)

        # Test with micro averaging
        metric_micro = ListF1Score(average="micro")
        score_micro = await metric_micro(y_true, y_pred)
        self.assertIsInstance(score_micro, float)

        # Test with macro averaging
        metric_macro = ListF1Score(average="macro")
        score_macro = await metric_macro(y_true, y_pred)
        self.assertIsInstance(score_macro, float)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.datasets.gsm8k import get_input_data_model as get_input_data_model
from synalinks.src.datasets.gsm8k import get_output_data_model as get_output_data_model
from synalinks.src.datasets.gsm8k import load_data as load_data



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.modules import SelfCritique
from synalinks.src.programs import Program
from synalinks.src.rewards.reward_wrappers import ProgramAsJudge
from synalinks.src.saving import serialization_lib


class LMAsJudgeProgram(Program):
    """Evaluate the output of a program using a `LanguageModel`.

    Args:
        language_model (LanguageModel): The language model to use.
        prompt_template (str): The default jinja2 prompt template
            to use (see `Generator`).
        examples (list): The default examples to use in the prompt
            (see `Generator`).
        instructions (list): The default instructions to use (see `Generator`).
        name (str): Optional. The name of the program.
        description (str): Optional. The description of the program.
        trainable (bool): Whether the program's variables should be trainable.
    """

    def __init__(
        self,
        language_model=None,
        prompt_template=None,
        examples=None,
        instructions=None,
        name=None,
        description=None,
        trainable=True,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        self.critique = SelfCritique(
            language_model=language_model,
            prompt_template=prompt_template,
            examples=examples,
            instructions=instructions,
            name="self_critique_" + self.name,
        )
        self.language_model = language_model
        self.prompt_template = prompt_template
        self.examples = examples
        self.instructions = instructions

    async def call(self, inputs):
        if not isinstance(inputs, (list, tuple)):
            raise ValueError("The inputs should be a list or tuple.")
        if len(inputs) != 2:
            raise ValueError("The inputs of the program should have a length of 2.")
        y_true = inputs[0]
        y_pred = inputs[1]
        if not y_pred:
            return 0.0
        if y_true:
            y_true = await ops.prefix(
                y_true,
                prefix="gold",
                name="gold_y_true",
            )
            return await self.critique(
                await ops.concat(
                    y_true,
                    y_pred,
                    name="y_true_with_y_pred",
                )
            )
        else:
            return await self.critique(y_pred)

    def get_config(self):
        config = {
            "prompt_template": self.prompt_template,
            "examples": self.examples,
            "instructions": self.instructions,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model
            )
        }
        return {**language_model_config, **config}

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model")
        )
        return cls(language_model=language_model, **config)


@synalinks_export(
    [
        "synalinks.LMAsJudge",
        "synalinks.rewards.LMAsJudge",
    ]
)
class LMAsJudge(ProgramAsJudge):
    """Evaluate the output of a program using a `LanguageModel`.

    Example:

    ```python

    async def main():
        # ... program definition

        program.compile(
            reward=synalinks.rewards.LMAsJudge(
                language_model=language_model,
            )
            optimizer=synalinks.optimizers.RandomFewShot(),
        )

        history = await program.fit(...)

    ```

    Args:
        language_model (LanguageModel): The language model to use.
        prompt_template (str): The default jinja2 prompt template
            to use (see `Generator`).
        instructions (list): The default instructions to use (see `Generator`).
        examples (list): The default examples to use in the prompt
            (see `Generator`).
        name (str): Optional. string name of the reward instance.
        in_mask (list): Optional. list of keys to keep to compute the reward.
        out_mask (list): Optional. list of keys to remove to compute the reward.
    """

    def __init__(
        self,
        language_model=None,
        prompt_template=None,
        examples=None,
        instructions=None,
        name="lm_as_judge",
        in_mask=None,
        out_mask=None,
    ):
        program = LMAsJudgeProgram(
            language_model=language_model,
            prompt_template=prompt_template,
            examples=examples,
            instructions=instructions,
        )
        super().__init__(
            program=program,
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )
